import logging
import torch
import torch.nn as nn
from gardening_tools.modules.networks.components.blocks import (
    MultiLayerConvDropoutNormNonlin,
)
from gardening_tools.modules.networks.components.heads import ClsRegHead
from gardening_tools.modules.networks.BaseNet import BaseNet
from gardening_tools.modules.networks.components.decoders import UNetDecoder
from gardening_tools.modules.networks.components.encoders import UNetEncoder


class UNet(BaseNet):
    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        encoder: nn.Module = UNetEncoder,
        encoder_basic_block=MultiLayerConvDropoutNormNonlin.get_block_constructor(2),
        decoder: nn.Module = UNetDecoder,
        decoder_basic_block=MultiLayerConvDropoutNormNonlin.get_block_constructor(2),
        dimensions: str = "3D",
        starting_filters: int = 32,
        use_skip_connections: bool = True,
        deep_supervision: bool = False,
    ):
        super().__init__()
        self.stem_weight_name = None
        if dimensions == "2D":
            conv_op = nn.Conv2d
            dropout_op = nn.Dropout2d
            norm_op = nn.InstanceNorm2d
            pool_op = nn.MaxPool2d
            upsample_op = torch.nn.ConvTranspose2d
        elif dimensions == "3D":
            conv_op = nn.Conv3d
            dropout_op = nn.Dropout3d
            norm_op = nn.InstanceNorm3d
            pool_op = nn.MaxPool3d
            upsample_op = torch.nn.ConvTranspose3d
        else:
            logging.warn("Uuh, dimensions not in ['2D', '3D']")

        self.encoder = encoder(
            basic_block=encoder_basic_block,
            conv_op=conv_op,
            dropout_op=dropout_op,
            input_channels=input_channels,
            norm_op=norm_op,
            pool_op=pool_op,
            starting_filters=starting_filters,
        )
        self.decoder = decoder(
            basic_block=decoder_basic_block,
            conv_op=conv_op,
            dropout_op=dropout_op,
            norm_op=norm_op,
            output_channels=output_channels,
            starting_filters=starting_filters,
            upsample_op=upsample_op,
            use_skip_connections=use_skip_connections,
            deep_supervision=deep_supervision,
        )
        self.num_classes = output_channels

    def forward(self, x):
        enc = self.encoder(x)
        return self.decoder(enc)


class UNetCLSREG(BaseNet):
    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        encoder: nn.Module = UNetEncoder,
        encoder_basic_block=MultiLayerConvDropoutNormNonlin.get_block_constructor(2),
        decoder: nn.Module = ClsRegHead,
        dimensions: str = "3D",
        starting_filters: int = 32,
        deep_supervision: bool = False,
    ):
        super().__init__()
        assert deep_supervision is False, "UNetCLSREG does not support deep supervision"
        self.stem_weight_name = None

        if dimensions == "2D":
            conv_op = nn.Conv2d
            dropout_op = nn.Dropout2d
            norm_op = nn.InstanceNorm2d
            pool_op = nn.MaxPool2d
            clsreg_pool_op = nn.AdaptiveAvgPool2d
        elif dimensions == "3D":
            conv_op = nn.Conv3d
            dropout_op = nn.Dropout3d
            norm_op = nn.InstanceNorm3d
            pool_op = nn.MaxPool3d
            clsreg_pool_op = nn.AdaptiveAvgPool3d
        else:
            logging.warn("Uuh, dimensions not in ['2D', '3D']")

        self.encoder = encoder(
            basic_block=encoder_basic_block,
            conv_op=conv_op,
            dropout_op=dropout_op,
            input_channels=input_channels,
            norm_op=norm_op,
            pool_op=pool_op,
            starting_filters=starting_filters,
        )
        self.decoder = decoder(
            pool_op=clsreg_pool_op,
            input_channels=starting_filters * 16,
            output_channels=output_channels,
            dropout_rate=0.2,
        )
        self.num_classes = output_channels

    def forward(self, x):
        enc = self.encoder(x)
        return self.decoder(enc)



import torch
import torch.nn as nn


class ClsRegHead(nn.Module):
    def __init__(self, input_channels, output_channels, pool_op, dropout_rate=0.0):
        super().__init__()
        if issubclass(pool_op, nn.AdaptiveAvgPool3d):
            self.global_pool = pool_op((1, 1, 1))
        elif issubclass(pool_op, nn.AdaptiveAvgPool2d):
            self.global_pool = pool_op((1, 1))
        self.fc = nn.Linear(input_channels, output_channels)
        self.drop = nn.Dropout(dropout_rate) if dropout_rate > 0 else nn.Identity()

    def forward(self, x):
        x = x[-1]
        x = self.global_pool(x)
        x = self.drop(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x



from torch import Tensor, argmax
from torchmetrics.segmentation import GeneralizedDiceScore as TorchGeneralizedDiceScore


class GeneralizedDiceScore(TorchGeneralizedDiceScore):
    """
    Generalized Dice Score metric adapted to accept logits as input (preds) and target of any dtype.
    """

    def update(self, preds: Tensor, target: Tensor) -> None:
        preds = argmax(preds, dim=1)
        return super().update(preds, target.long())



import torch


def volume_wise_znorm(x: torch.Tensor) -> torch.Tensor:
    """
    Normalize the input tensor using volume-wise z-normalization.
    Assumes the input is a 3D tensor (X, Y, Z) or 2D Tensor (X, Y).
    """
    mean = x.mean()
    std = x.std(unbiased=False).clamp_min(1e-8)
    return (x - mean) / std


def ct_normalization(x: torch.Tensor) -> torch.Tensor:
    """
    Normalize the input tensor using CT-specific normalization.
    Assumes the input is a 3D tensor (X, Y, Z) or 2D Tensor (X, Y).
    """
    x = x.clamp(-1024, 1024) / 1024
    return x



import math
import numpy as np
import torch
from timm.layers import (
    PatchDropout,
    RotaryEmbeddingCat,
    apply_keep_indices_nlc,
    trunc_normal_,
)
from timm.models.eva import EvaBlock
from torch import nn
from torch.nn import LayerNorm
from torch.utils.checkpoint import checkpoint
from typing import Callable, Optional, Tuple


class Eva(nn.Module):
    """Eva Vision Transformer w/ Abs & Rotary Pos Embed

    This class implements the EVA and EVA02 models that were based on the BEiT ViT variant
      * EVA - abs pos embed, global avg pool
      * EVA02 - abs + rope pos embed, global avg pool, SwiGLU, scale Norm in MLP (ala normformer)


    """

    def __init__(
        self,
        embed_dim: int = 1024,
        depth: int = 24,
        num_heads: int = 16,
        qkv_bias: bool = True,
        qkv_fused: bool = False,
        mlp_ratio: float = 4 * 2 / 3,
        swiglu_mlp: bool = True,
        scale_mlp: bool = True,
        scale_attn_inner: bool = False,
        pos_drop_rate: float = 0.0,
        patch_drop_rate: float = 0.0,  # drops input patches, may be used for MAE style pretraining
        proj_drop_rate: float = 0.0,  # drops out things related to the projection. That is in the MLP and at the end of \
        # EVA attention
        attn_drop_rate: float = 0.0,  # drops attention, meaning connections between patches may bebroken up at random
        drop_path_rate: float = 0.0,  # drops computations (multihead attention, mlp), Implementation of scaling might be \
        # useless here because this is not batch normed
        norm_layer: Callable = LayerNorm,
        init_values: Optional[float] = None,
        use_abs_pos_emb: bool = True,
        use_rot_pos_emb: bool = True,
        dynamic_img_size: bool = False,
        ref_feat_shape: Optional[Tuple[int, ...]] = None,  # 224/14
        num_reg_tokens: int = 0,
        rope_impl=RotaryEmbeddingCat,
        rope_kwargs=None,
        block_fn=EvaBlock,
    ):
        """
        Diff to timm implementation

        - removed patch embedding, we expect embeded patches
        - removed classification token, we use features at the end
        - removed head
        - dynamic image size is not supported, but left in for future stuff
        - self.cls_token removed
        - removed postnorm block support
        """
        super().__init__()
        if rope_kwargs is None:
            rope_kwargs = {}

        self.num_features = self.embed_dim = (
            embed_dim  # num_features for consistency with other models
        )
        self.dynamic_img_size = dynamic_img_size
        self.grad_checkpointing = False

        self.num_prefix_tokens = num_reg_tokens

        num_patches = np.prod(ref_feat_shape)

        self.pos_embed = (
            nn.Parameter(
                torch.zeros(1, num_patches + self.num_prefix_tokens, embed_dim)
            )
            if use_abs_pos_emb
            else None
        )
        self.pos_drop = nn.Dropout(p=pos_drop_rate)
        if patch_drop_rate > 0:
            self.patch_drop = PatchDropout(
                patch_drop_rate,
                num_prefix_tokens=self.num_prefix_tokens,
            )
        else:
            self.patch_drop = None

        if use_rot_pos_emb:
            if len(ref_feat_shape) == 3:
                rope_dim = round(embed_dim // num_heads / 1.5)
                assert rope_dim == embed_dim / num_heads / 1.5, (
                    "rope dim must be divsible by (num_heads * 1.5)"
                )
                assert rope_dim % 4 == 0, "rope dim must be divisible by 4"
            else:
                rope_dim = embed_dim // num_heads
            self.rope = rope_impl(
                rope_dim,
                in_pixels=False,
                feat_shape=ref_feat_shape,
                ref_feat_shape=ref_feat_shape,
                **rope_kwargs,
            )
        else:
            self.rope = None

        dpr = [
            x.item() for x in torch.linspace(0, drop_path_rate, depth)
        ]  # stochastic depth decay rule
        block_fn = block_fn
        self.blocks = nn.ModuleList(
            [
                block_fn(
                    dim=embed_dim,
                    num_heads=num_heads,
                    qkv_bias=qkv_bias,
                    qkv_fused=qkv_fused,
                    mlp_ratio=mlp_ratio,
                    swiglu_mlp=swiglu_mlp,
                    scale_mlp=scale_mlp,
                    scale_attn_inner=scale_attn_inner,
                    proj_drop=proj_drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[i],
                    norm_layer=norm_layer,
                    init_values=init_values,
                    num_prefix_tokens=self.num_prefix_tokens,
                )
                for i in range(depth)
            ]
        )

        self.norm = norm_layer(embed_dim)

        self.apply(self._init_weights)
        if self.pos_embed is not None:
            trunc_normal_(self.pos_embed, std=0.02)

        self.fix_init_weight()

    def fix_init_weight(self):
        def rescale(param, layer_id):
            param.div_(math.sqrt(2.0 * layer_id))

        for layer_id, layer in enumerate(self.blocks):
            rescale(layer.attn.proj.weight.data, layer_id + 1)
            rescale(layer.mlp.fc2.weight.data, layer_id + 1)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.zeros_(m.bias)

    @torch.jit.ignore
    def no_weight_decay(self):
        nwd = {"pos_embed", "cls_token"}
        return nwd

    @torch.jit.ignore
    def set_grad_checkpointing(self, enable=True):
        self.grad_checkpointing = enable

    @torch.jit.ignore
    def group_matcher(self, coarse=False):
        matcher = dict(
            stem=r"^cls_token|pos_embed|patch_embed",  # stem and embed
            blocks=[(r"^blocks\.(\d+)", None), (r"^norm", (99999,))],
        )
        return matcher

    def _pos_embed(self, x) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        if self.dynamic_img_size:
            raise NotImplementedError(
                "dynamic_img_size is not implemented at the moment"
            )
        else:
            pos_embed = self.pos_embed
            rot_pos_embed = self.rope.get_embed() if self.rope is not None else None

        if pos_embed is not None:
            x = x + pos_embed
        x = self.pos_drop(x)

        # obtain shared rotary position embedding and apply patch dropout
        if self.patch_drop is not None:
            x, keep_indices = self.patch_drop(x)
            if rot_pos_embed is not None and keep_indices is not None:
                rot_pos_embed = apply_keep_indices_nlc(x, rot_pos_embed, keep_indices)
            return x, rot_pos_embed, keep_indices
        else:
            return x, rot_pos_embed, None

    def forward_features(self, x):
        x, rot_pos_embed, keep_indices = self._pos_embed(x)
        for blk in self.blocks:
            if self.grad_checkpointing and not torch.jit.is_scripting():
                x = checkpoint(blk, x, rope=rot_pos_embed)
            else:
                x = blk(x, rope=rot_pos_embed)
        x = self.norm(x)
        return x, keep_indices

    def forward(self, x):
        x, keep_indices = self.forward_features(x)
        return x, keep_indices



from gardening_tools.functional.transforms.sampling import torch_downsample_label
from gardening_tools.modules.transforms.BaseTransform import BaseTransform


class Torch_DownsampleSegForDS(BaseTransform):
    def __init__(
        self,
        deep_supervision: bool = False,
        label_key="label",
        factors=(1, 1 / 2, 1 / 4, 1 / 8, 1 / 16, 1 / 16),
    ):
        self.deep_supervision = deep_supervision
        self.label_key = label_key
        self.factors = factors

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __downsample__(self, label, factors):
        downsampled_labels = torch_downsample_label(label, factors)
        return downsampled_labels

    def __call__(self, data_dict):
        if self.deep_supervision:
            data_dict[self.label_key] = self.__downsample__(
                data_dict[self.label_key], self.factors
            )
        return data_dict



import torch
import torch.nn as nn
from gardening_tools.modules.networks.components.blocks import (
    DoubleConvDropoutNormNonlin,
    MultiLayerConvDropoutNormNonlin,
    ResidualBlock,
    StackedResidualBlocks,
)
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.dropout import _DropoutNd
from typing import List, Tuple, Type, Union


class UNetEncoder(nn.Module):
    def __init__(
        self,
        input_channels: int,
        basic_block=DoubleConvDropoutNormNonlin,
        conv_op=nn.Conv3d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "bias": True,
        },
        norm_op=nn.InstanceNorm3d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout3d,
        dropout_op_kwargs={"p": 0.0, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
        weightInitializer=None,
        pool_op=nn.MaxPool3d,
        starting_filters: int = 64,
    ) -> None:
        super().__init__()

        # Task specific
        self.filters = starting_filters

        # Model parameters
        self.basic_block = basic_block
        self.conv_op = conv_op
        self.conv_kwargs = conv_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.norm_op = norm_op
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.weightInitializer = weightInitializer
        self.pool_op = pool_op

        self.in_conv = self.basic_block(
            input_channels=input_channels,
            output_channels=self.filters,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.pool1 = self.pool_op(2)
        self.encoder_conv1 = self.basic_block(
            input_channels=self.filters,
            output_channels=self.filters * 2,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.pool2 = self.pool_op(2)
        self.encoder_conv2 = self.basic_block(
            input_channels=self.filters * 2,
            output_channels=self.filters * 4,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.pool3 = self.pool_op(2)
        self.encoder_conv3 = self.basic_block(
            input_channels=self.filters * 4,
            output_channels=self.filters * 8,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.pool4 = self.pool_op(2)
        self.encoder_conv4 = self.basic_block(
            input_channels=self.filters * 8,
            output_channels=self.filters * 16,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        if self.weightInitializer is not None:
            print("initializing weights")
            self.apply(self.weightInitializer)

    def forward(self, x):
        x0 = self.in_conv(x)

        x1 = self.pool1(x0)
        x1 = self.encoder_conv1(x1)

        x2 = self.pool2(x1)
        x2 = self.encoder_conv2(x2)

        x3 = self.pool3(x2)
        x3 = self.encoder_conv3(x3)

        x4 = self.pool4(x3)
        x4 = self.encoder_conv4(x4)

        return [x0, x1, x2, x3, x4]


class ResidualUNetEncoder(nn.Module):
    def __init__(
        self,
        input_channels: int,
        features_per_stage: Union[int, List[int], Tuple[int, ...]],
        conv_op: Type[_ConvNd],
        pool_op: Union[nn.AvgPool2d, nn.AvgPool3d],
        kernel_size: int,
        stride: int,
        n_blocks_per_stage: Union[int, List[int], Tuple[int, ...]],
        conv_bias: bool = True,
        norm_op: Union[None, Type[nn.Module]] = None,
        norm_op_kwargs: dict = None,
        dropout_op: Union[None, Type[_DropoutNd]] = None,
        dropout_op_kwargs: dict = None,
        nonlin: Union[None, Type[torch.nn.Module]] = None,
        nonlin_kwargs: dict = None,
        block: Type[ResidualBlock] = ResidualBlock,
    ):
        super().__init__()

        stem_channels = features_per_stage[0]

        self.stem = MultiLayerConvDropoutNormNonlin(
            input_channels=input_channels,
            output_channels=stem_channels,
            num_layers=1,
            conv_op=conv_op,
            conv_kwargs={
                "kernel_size": kernel_size,
                "stride": 1,
                "bias": conv_bias,
            },
            norm_op=norm_op,
            norm_op_kwargs=norm_op_kwargs,
            dropout_op=dropout_op,
            dropout_op_kwargs=dropout_op_kwargs,
            nonlin=nonlin,
            nonlin_kwargs=nonlin_kwargs,
        )

        input_channels = stem_channels

        stages = []
        for s in range(len(features_per_stage)):
            stage = StackedResidualBlocks(
                n_blocks=n_blocks_per_stage[s],
                conv_op=conv_op,
                input_channels=input_channels,
                output_channels=features_per_stage[s],
                kernel_size=kernel_size,
                initial_stride=stride if s != 0 else 1,
                conv_bias=conv_bias,
                norm_op=norm_op,
                norm_op_kwargs=norm_op_kwargs,
                dropout_op=dropout_op,
                dropout_op_kwargs=dropout_op_kwargs,
                nonlin=nonlin,
                nonlin_kwargs=nonlin_kwargs,
                block=block,
                pool_op=pool_op,
            )

            stages.append(stage)
            input_channels = features_per_stage[s]

        self.stages = nn.Sequential(*stages)

    def forward(self, x):
        if self.stem is not None:
            x = self.stem(x)
        ret = []
        for s in self.stages:
            x = s(x)
            ret.append(x)
        return ret



from gardening_tools.modules.transforms.BaseTransform import BaseTransform
from gardening_tools.functional.transforms.ringing import (
    numpy_gibbs_ringing,
    torch_gibbs_ringing,
)
import numpy as np


class Numpy_GibbsRinging(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        cut_freq=(96, 129),
        axes=(0, 3),
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.cut_freq = cut_freq
        self.axes = axes
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(cut_freq, axes):
        cut_freq = np.random.randint(*cut_freq)
        axis = np.random.randint(*axes)
        return cut_freq, axis

    def __gibbsRinging__(self, image, num_sample, axis):
        image = numpy_gibbs_ringing(
            image,
            num_sample=num_sample,
            axis=axis,
            clip_to_input_range=self.clip_to_input_range,
        )
        return image

    def __call__(self, data_dict):
        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    cut_freq, axis = self.get_params(self.cut_freq, self.axes)
                    data_dict[self.data_key][b, c] = self.__gibbsRinging__(
                        data_dict[self.data_key][b, c], cut_freq, axis
                    )
        return data_dict


class Torch_GibbsRinging(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_per_channel: float = 0.0,
        cut_freq=(96, 129),
        axes=(0, 3),
        clip_to_input_range=False,
        batched: bool = True,
    ):
        self.data_key = data_key
        self.p_per_channel = p_per_channel
        self.cut_freq = cut_freq
        self.axes = axes
        self.clip_to_input_range = clip_to_input_range
        self.batched = batched

    @staticmethod
    def get_params(cut_freq, axes):
        cut_freq = np.random.randint(*cut_freq)
        axis = np.random.randint(*axes)
        return cut_freq, axis

    def __gibbsRinging__(self, image, num_sample, axis):
        image = torch_gibbs_ringing(
            image,
            num_sample=num_sample,
            axes=[axis],
            clip_to_input_range=self.clip_to_input_range,
        )
        return image

    def __call__(self, data_dict):
        if not self.batched:
            for c in range(data_dict[self.data_key].shape[0]):
                if np.random.uniform() < self.p_per_channel:
                    cut_freq, axis = self.get_params(self.cut_freq, self.axes)
                    data_dict[self.data_key][c] = self.__gibbsRinging__(
                        data_dict[self.data_key][c], cut_freq, axis
                    )
        else:
            for b in range(data_dict[self.data_key].shape[0]):
                for c in range(data_dict[self.data_key].shape[1]):
                    if np.random.uniform() < self.p_per_channel:
                        cut_freq, axis = self.get_params(self.cut_freq, self.axes)
                        data_dict[self.data_key][b, c] = self.__gibbsRinging__(
                            data_dict[self.data_key][b, c], cut_freq, axis
                        )

        return data_dict



from gardening_tools.modules.transforms.BaseTransform import BaseTransform
import numpy as np
from gardening_tools.functional.transforms.bias_field import (
    numpy_bias_field,
    torch_bias_field,
)


class Numpy_BiasField(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __biasField__(self, image):
        image = numpy_bias_field(image, clip_to_input_range=self.clip_to_input_range)
        return image

    def __call__(self, data_dict):
        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    data_dict[self.data_key][b, c] = self.__biasField__(
                        data_dict[self.data_key][b, c]
                    )
        return data_dict


class Torch_BiasField(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_per_channel: float = 0.0,
        clip_to_input_range=False,
        batched: bool = True,
    ):
        self.data_key = data_key
        self.p_per_channel = p_per_channel
        self.clip_to_input_range = clip_to_input_range
        self.batched = batched

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __biasField__(self, image):
        image = torch_bias_field(image, clip_to_input_range=self.clip_to_input_range)
        return image

    def __call__(self, data_dict):
        if not self.batched:
            for c in range(data_dict[self.data_key].shape[0]):
                if np.random.uniform() < self.p_per_channel:
                    data_dict[self.data_key][c] = self.__biasField__(
                        data_dict[self.data_key][c]
                    )
        else:
            for b in range(data_dict[self.data_key].shape[0]):
                for c in range(data_dict[self.data_key].shape[1]):
                    if np.random.uniform() < self.p_per_channel:
                        data_dict[self.data_key][b, c] = self.__biasField__(
                            data_dict[self.data_key][b, c]
                        )
        return data_dict



from gardening_tools.modules.transforms.BaseTransform import BaseTransform
from gardening_tools.functional.normalization import volume_wise_znorm, ct_normalization


class Torch_Normalize(BaseTransform):
    def __init__(
        self, normalize: bool = False, data_key: str = "image", fn=volume_wise_znorm
    ):
        self.normalize = normalize
        self.data_key = data_key
        self.fn = fn

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __normalize__(self, data_dict):
        for c in range(data_dict[self.data_key].shape[0]):
            if isinstance(self.fn, list):
                fn = self.fn[c]
            else:
                fn = self.fn
            data_dict[self.data_key][c] = fn(data_dict[self.data_key][c])
        return data_dict

    def __call__(self, data_dict):
        if self.normalize:
            data_dict = self.__normalize__(data_dict)
        return data_dict


class Torch_CT_NormalizeC0(BaseTransform):
    def __init__(
        self,
        normalize: bool = False,
        data_key: str = "image",
        non_ct_fn=volume_wise_znorm,
        ct_fn=ct_normalization,
    ):
        self.normalize = normalize
        self.data_key = data_key
        self.ct_fn = ct_fn
        self.non_ct_fn = non_ct_fn

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __normalize__(self, data_dict):
        for c in range(data_dict[self.data_key].shape[0]):
            if c == 0:
                fn = self.ct_fn
            else:
                fn = self.non_ct_fn
            data_dict[self.data_key][c] = fn(data_dict[self.data_key][c])
        return data_dict

    def __call__(self, data_dict):
        if self.normalize:
            data_dict = self.__normalize__(data_dict)
        return data_dict



import numpy as np


def dice(tp, fp, tn, fn):
    try:
        return (2 * tp) / (2 * tp + fp + fn)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def dice_per_label(tp_list, fp_list, tn_list, fn_list):
    return [
        dice(tp_list[i], fp_list[i], tn_list[i], fn_list[i])
        for i in range(len(tp_list))
    ]


def jaccard(tp, fp, tn, fn):
    try:
        return (tp) / (tp + fp + fn)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def jaccard_per_label(tp_list, fp_list, tn_list, fn_list):
    return [
        jaccard(tp_list[i], fp_list[i], tn_list[i], fn_list[i])
        for i in range(len(tp_list))
    ]


def sensitivity(tp, fp, tn, fn):
    try:
        return tp / (tp + fn)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def specificity(tp, fp, tn, fn):
    try:
        return tn / (tn + fp)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def precision(tp, fp, tn, fn):
    try:
        return tp / (tp + fp)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def volume_similarity(tp, fp, tn, fn):
    try:
        return 1 - abs(fn - fp) / (2 * tp + fn + fp)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def f1(tp, fp, tn, fn):
    try:
        prec = precision(tp, fp, tn, fn)
        sens = sensitivity(tp, fp, tn, fn)
        return 2 * (prec * sens) / (prec + sens)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def accuracy(tp, fp, tn, fn):
    try:
        return (tp + tn) / (tp + fp + tn + fn)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def TP(tp, fp, tn, fn):
    return tp


def FP(tp, fp, tn, fn):
    return fp


def FN(tp, fp, tn, fn):
    return fn


def total_pos_gt(tp, fp, tn, fn):
    return tp + fn


def total_pos_pred(tp, fp, tn, fn):
    return tp + fp



import numpy as np
import torch
import torch.nn.functional as F
from typing import Sequence, List
from skimage.transform import resize


def torch_downsample_label(
    label: torch.Tensor, factors: Sequence[float]
) -> List[torch.Tensor]:
    """
    Nearest-neighbour down-sampling for label volumes stored as **(B, C, X, Y, Z)**.

    Parameters
    ----------
    label : torch.Tensor
        Input tensor of shape *(B, C, X, Y, Z)*. Any floating-, integer-, or
        boolean-dtype tensor is accepted.
    factors : Sequence[float]
        Iterable of scale factors.  Each factor is applied uniformly to the
        three spatial axes (X, Y, Z).  A factor of 1 returns the original
        tensor instance.

    Returns
    -------
    List[torch.Tensor]
        One tensor per factor, each shaped *(B, C, X', Y', Z')* with the same
        dtype as the input.  Spatial extents are truncated to `floor(dim × f)`.
    """
    if label.ndim != 5:
        raise ValueError("Expects (B,C,H,W,D) tensor, got {}".format(label.shape))
    orig_dtype = label.dtype
    outs = []
    for f in factors:
        if f == 1:
            outs.append(label)
            continue
        scaled = F.interpolate(
            label.float(),
            scale_factor=[f, f, f],
            mode="nearest-exact",  # closest to numpys "edge"-mode, with recompute below
            recompute_scale_factor=True,  # this is needed to reproduce numpy version
        )
        restored = scaled.to(orig_dtype)
        outs.append(restored)
    return outs


def torch_resize(
    image: torch.Tensor,
    target_size: tuple[int],
    clip_to_input_range: bool,
) -> torch.Tensor:
    img_min = image.min()
    img_max = image.max()

    image = image.float()

    if image.ndim == 3:
        mode = "bicubic"
    elif image.ndim == 4:
        if len(target_size) == 2:
            mode = "bicubic"
            image = image[:, torch.randint(0, image.shape[1], (1,)).item()]
        else:
            mode = "trilinear"
    else:
        raise ValueError("Image must be 3D or 4D.")

    result = F.interpolate(
        image.unsqueeze(0), size=tuple(target_size), mode=mode
    ).squeeze(0)

    if clip_to_input_range:
        result = result.clamp(min=img_min.item(), max=img_max.item())

    return result


def numpy_simulate_lowres(image, target_shape, clip_to_input_range):
    img_min = image.min()
    img_max = image.max()
    shape = image.shape
    downsampled = resize(
        image.astype(float),
        target_shape,
        order=0,
        mode="edge",
        anti_aliasing=False,
    )
    image = resize(downsampled, shape, order=3, mode="edge", anti_aliasing=False)
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)
    return image


def torch_simulate_lowres(
    image: torch.Tensor,
    target_shape: tuple[int],
    clip_to_input_range: bool,
) -> torch.Tensor:
    original_shape = image.shape
    img_min = image.min()
    img_max = image.max()

    image = image.float()
    if image.ndim == 2:
        image = image.unsqueeze(0).unsqueeze(0)
        mode_down = "nearest-exact"
        mode_up = "bicubic"
    elif image.ndim == 3:
        image = image.unsqueeze(0).unsqueeze(0)
        mode_down = "nearest-exact"
        mode_up = "trilinear"
    else:
        raise ValueError("Image must be 3D or 4D.")

    downsampled = F.interpolate(image, size=tuple(target_shape), mode=mode_down)
    upsampled = F.interpolate(
        downsampled, size=tuple(original_shape), mode=mode_up, align_corners=False
    )
    result = upsampled.squeeze(0).squeeze(0)

    if clip_to_input_range:
        result = result.clamp(min=img_min.item(), max=img_max.item())

    return result



import torch
import numpy as np
import torch.nn.functional as F
from torch import nn


def sum_tensor(inp, axes, keepdim=False):
    axes = np.unique(axes).astype(int)
    if keepdim:
        for ax in axes:
            inp = inp.sum(int(ax), keepdim=True)
    else:
        for ax in sorted(axes, reverse=True):
            inp = inp.sum(int(ax))
    return inp


def get_tp_fp_fn_tn(net_output, gt, axes=None, mask=None, square=False):
    """
    net_output must be (b, c, x, y(, z)))
    gt must be a label map (shape (b, 1, x, y(, z)) OR shape (b, x, y(, z)))
        or one hot encoding (b, c, x, y(, z))
    if mask is provided it must have shape (b, 1, x, y(, z)))
    :param net_output:
    :param gt:
    :param axes: can be (, ) = no summation
    :param mask: mask must be 1 for valid pixels and 0 for invalid pixels
    :param square: if True then fp, tp and fn will be squared before summation
    :return:
    """

    if axes is None:
        axes = tuple(range(2, len(net_output.size())))

    shp_x = net_output.shape
    shp_y = gt.shape
    with torch.no_grad():
        if len(shp_x) != len(shp_y):
            gt = gt.view((shp_y[0], 1, *shp_y[1:]))

        if all([i == j for i, j in zip(net_output.shape, gt.shape)]):
            # if this is the case then gt is probably already a one hot encoding
            y_onehot = gt
        else:
            gt = gt.long()
            y_onehot = torch.zeros(shp_x).to(net_output)
            y_onehot.scatter_(1, gt, 1)

    tp = net_output * y_onehot
    fp = net_output * (1 - y_onehot)
    fn = (1 - net_output) * y_onehot
    tn = (1 - net_output) * (1 - y_onehot)

    if mask is not None:
        tp = torch.stack(
            tuple(x_i * mask[:, 0] for x_i in torch.unbind(tp, dim=1)), dim=1
        )
        fp = torch.stack(
            tuple(x_i * mask[:, 0] for x_i in torch.unbind(fp, dim=1)), dim=1
        )
        fn = torch.stack(
            tuple(x_i * mask[:, 0] for x_i in torch.unbind(fn, dim=1)), dim=1
        )
        tn = torch.stack(
            tuple(x_i * mask[:, 0] for x_i in torch.unbind(tn, dim=1)), dim=1
        )

    if square:
        tp = tp**2
        fp = fp**2
        fn = fn**2
        tn = tn**2

    if len(axes) > 0:
        tp = sum_tensor(tp, axes, keepdim=False)
        fp = sum_tensor(fp, axes, keepdim=False)
        fn = sum_tensor(fn, axes, keepdim=False)
        tn = sum_tensor(tn, axes, keepdim=False)

    return tp, fp, fn, tn


class SoftDiceLoss(nn.Module):
    def __init__(self, apply_softmax=True, batch_dice=False, do_bg=True, smooth=1.0):
        """ """
        super(SoftDiceLoss, self).__init__()

        self.do_bg = do_bg
        self.batch_dice = batch_dice
        self.smooth = smooth
        self.apply_softmax = apply_softmax

    def forward(self, x, y, loss_mask=None):
        shp_x = x.shape

        if self.batch_dice:
            axes = [0] + list(range(2, len(shp_x)))
        else:
            axes = list(range(2, len(shp_x)))

        if self.apply_softmax is True:
            x = F.softmax(x, 1)

        tp, fp, fn, _ = get_tp_fp_fn_tn(x, y, axes, loss_mask, False)

        nominator = 2 * tp + self.smooth
        denominator = 2 * tp + fp + fn + self.smooth

        dc = nominator / (denominator + 1e-8)

        if not self.do_bg:
            if self.batch_dice:
                dc = dc[1:]
            else:
                dc = dc[:, 1:]
        dc = dc.mean()

        return -dc



import torch
from gardening_tools.modules.transforms.BaseTransform import BaseTransform


class Torch_Clamp(BaseTransform):
    def __init__(
        self,
        clamp: bool = False,
        data_key: str = "image",
        lower_percentile: float = 0.01,
        upper_percentile: float = 0.99,
    ):
        """
        Initialize the Torch_Clamp transform.

        Args:
            clamp (bool): Whether to apply clamping
            data_key (str): Key of the data to clamp in the data dictionary
            lower_percentile (float): Lower percentile for clamping when data has negative values (default: 0.01)
            upper_percentile (float): Upper percentile for clamping (default: 0.99)
        """
        self.clamp = clamp
        self.data_key = data_key
        self.lower_percentile = lower_percentile
        self.upper_percentile = upper_percentile

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __clamp__(self, data_dict):
        for c in range(data_dict[self.data_key].shape[0]):
            channel_data = data_dict[self.data_key][c]

            if channel_data.ndim == 2:
                flat_data = channel_data[::2, ::2].flatten()
            elif channel_data.ndim == 3:
                flat_data = channel_data[::2, ::2, ::2].flatten()
            else:
                flat_data = channel_data.flatten()

            sorted_data = torch.sort(flat_data)[0]  # Sort only once
            n = sorted_data.numel()

            if self.lower_percentile is not None:
                lower_idx = int(self.lower_percentile * (n - 1))
                lower_bound = sorted_data[lower_idx]
            else:
                lower_bound = None

            if self.upper_percentile is not None:
                upper_idx = int(self.upper_percentile * (n - 1))
                upper_bound = sorted_data[upper_idx]
            else:
                upper_bound = None

            data_dict[self.data_key][c] = torch.clamp(
                channel_data, min=lower_bound, max=upper_bound
            )

        return data_dict

    def __call__(self, data_dict):
        if self.clamp:
            data_dict = self.__clamp__(data_dict)
        return data_dict



import torch
from gardening_tools.modules.transforms.BaseTransform import BaseTransform


class Torch_CastLabelToDtype(BaseTransform):
    def __init__(self, label_key, dtype):
        self.label_key = label_key
        self.dtype = dtype

    @staticmethod
    def get_params(dtype):
        if dtype == "int":
            return torch.int64
        elif dtype == "float":
            return torch.float32

    def __cast__(self, data_dict, target_dtype):
        data_dict[self.label_key] = data_dict[self.label_key].to(target_dtype)
        return data_dict

    def __call__(self, data_dict):
        dtype = self.get_params(self.dtype)
        data_dict = self.__cast__(data_dict, dtype)
        return data_dict



from gardening_tools.modules.transforms.BaseTransform import BaseTransform
from gardening_tools.functional.transforms.gamma import numpy_gamma, torch_gamma
import numpy as np


class Numpy_Gamma(BaseTransform):
    """
    WRAPPER FOR NNUNET AUGMENT GAMMA: https://github.com/MIC-DKFZ/batchgenerators/blob/8822a08a7dbfa4986db014e6a74b040778164ca6/batchgenerators/augmentations/color_augmentations.py
    """

    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        p_invert_image: float = 0.05,
        gamma_range=(0.5, 2.0),
        per_channel=True,
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.gamma_range = gamma_range
        self.p_invert_image = p_invert_image
        self.per_channel = per_channel
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(p_invert_image):
        # No parameters to retrieve
        do_invert = False
        if np.random.uniform() < p_invert_image:
            do_invert = True
        return do_invert

    def __gamma__(self, image, gamma_range, invert_image, per_channel):
        return numpy_gamma(
            image,
            gamma_range,
            invert_image,
            per_channel,
            clip_to_input_range=self.clip_to_input_range,
        )

    def __call__(self, data_dict):
        for b in range(data_dict[self.data_key].shape[0]):
            if np.random.uniform() < self.p_per_sample:
                do_invert = self.get_params(self.p_invert_image)
                data_dict[self.data_key][b] = self.__gamma__(
                    data_dict[self.data_key][b],
                    self.gamma_range,
                    do_invert,
                    per_channel=self.per_channel,
                )
        return data_dict


class Torch_Gamma(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_all_channel: float = 0.0,
        p_invert_image: float = 0.05,
        gamma_range=(0.5, 2.0),
        per_channel=True,
        clip_to_input_range=False,
        batched: bool = True,
    ):
        self.data_key = data_key
        self.p_all_channel = p_all_channel
        self.gamma_range = gamma_range
        self.p_invert_image = p_invert_image
        self.per_channel = per_channel
        self.clip_to_input_range = clip_to_input_range
        self.batched = batched

    @staticmethod
    def get_params(p_invert_image):
        # No parameters to retrieve
        do_invert = False
        if np.random.uniform() < p_invert_image:
            do_invert = True
        return do_invert

    def __gamma__(self, image, gamma_range, invert_image, per_channel):
        return torch_gamma(
            image,
            gamma_range,
            invert_image,
            per_channel,
            clip_to_input_range=self.clip_to_input_range,
        )

    def __call__(self, data_dict):
        if not self.batched:
            if np.random.uniform() < self.p_all_channel:
                do_invert = self.get_params(self.p_invert_image)
                data_dict[self.data_key] = self.__gamma__(
                    data_dict[self.data_key],
                    self.gamma_range,
                    do_invert,
                    per_channel=self.per_channel,
                )
        else:
            for b in range(data_dict[self.data_key].shape[0]):
                if np.random.uniform() < self.p_all_channel:
                    do_invert = self.get_params(self.p_invert_image)
                    data_dict[self.data_key][b] = self.__gamma__(
                        data_dict[self.data_key][b],
                        self.gamma_range,
                        do_invert,
                        per_channel=self.per_channel,
                    )
        return data_dict



import numpy as np
from gardening_tools.functional.transforms.motion_ghosting import (
    numpy_motion_ghosting,
    torch_motion_ghosting,
)
from gardening_tools.modules.transforms.BaseTransform import BaseTransform
from typing import Tuple


class Numpy_MotionGhosting(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        alpha=(0.85, 0.95),
        num_reps=(2, 5),
        axes=(0, 3),
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.alpha = alpha
        self.num_reps = num_reps
        self.axes = axes
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(
        alpha: Tuple[float], num_reps: Tuple[float], axes: Tuple[float]
    ) -> Tuple[float]:
        alpha = np.random.uniform(*alpha)
        num_reps = np.random.randint(*num_reps)
        axis = np.random.randint(*axes)
        return alpha, num_reps, axis

    def __motionGhosting__(self, image, alpha, num_reps, axis):
        image = numpy_motion_ghosting(
            image=image,
            alpha=alpha,
            num_reps=num_reps,
            axis=axis,
            clip_to_input_range=self.clip_to_input_range,
        )
        return image

    def __call__(self, data_dict):
        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    alpha, num_reps, axis = self.get_params(
                        self.alpha, self.num_reps, self.axes
                    )
                    data_dict[self.data_key][b, c] = self.__motionGhosting__(
                        data_dict[self.data_key][b, c], alpha, num_reps, axis
                    )
        return data_dict


class Torch_MotionGhosting(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_per_channel: float = 0.0,
        alpha=(0.85, 0.95),
        num_reps=(2, 5),
        axes=(0, 3),
        clip_to_input_range=False,
        batched: bool = True,
    ):
        self.data_key = data_key
        self.p_per_channel = p_per_channel
        self.alpha = alpha
        self.num_reps = num_reps
        self.axes = axes
        self.clip_to_input_range = clip_to_input_range
        self.batched = batched

    @staticmethod
    def get_params(
        alpha: Tuple[float], num_reps: Tuple[float], axes: Tuple[float]
    ) -> Tuple[float]:
        alpha = np.random.uniform(*alpha)
        num_reps = np.random.randint(*num_reps)
        axis = np.random.randint(*axes)
        return alpha, num_reps, axis

    def __motionGhosting__(self, image, alpha, num_reps, axis):
        image = torch_motion_ghosting(
            image=image,
            alpha=alpha,
            num_reps=num_reps,
            axis=axis,
            clip_to_input_range=self.clip_to_input_range,
        )
        return image

    def __call__(self, data_dict):
        if not self.batched:
            for c in range(data_dict[self.data_key].shape[0]):
                if np.random.uniform() < self.p_per_channel:
                    alpha, num_reps, axis = self.get_params(
                        self.alpha, self.num_reps, self.axes
                    )
                    data_dict[self.data_key][c] = self.__motionGhosting__(
                        data_dict[self.data_key][c], alpha, num_reps, axis
                    )
        else:
            for b in range(data_dict[self.data_key].shape[0]):
                for c in range(data_dict[self.data_key].shape[1]):
                    if np.random.uniform() < self.p_per_channel:
                        alpha, num_reps, axis = self.get_params(
                            self.alpha, self.num_reps, self.axes
                        )
                        data_dict[self.data_key][b, c] = self.__motionGhosting__(
                            data_dict[self.data_key][b, c], alpha, num_reps, axis
                        )
        return data_dict



import numpy as np
import torch
from gardening_tools.modules.transforms.BaseTransform import BaseTransform
from gardening_tools.functional.transforms.cropping_and_padding import (
    torch_croppad,
    numpy_croppad,
    fit_image_to_patch_size,
)
from typing import Union, Literal

__all__ = ["Numpy_CropPad", "Torch_CropPad", "Numpy_Pad", "Torch_Pad"]


class Numpy_CropPad(BaseTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        pad_value: Union[Literal["min", "zero", "edge"], int, float] = "min",
        patch_size: tuple | list = None,
        p_oversample_foreground=0.0,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.pad_value = pad_value
        self.patch_size = patch_size
        self.p_oversample_foreground = p_oversample_foreground

    @staticmethod
    def get_params(data, pad_value, target_shape):
        if pad_value == "min":
            pad_kwargs = {"constant_values": data.min(), "mode": "constant"}
        elif pad_value == "zero":
            pad_kwargs = {
                "constant_values": np.zeros(1, dtype=data.dtype),
                "mode": "constant",
            }
        elif isinstance(pad_value, int) or isinstance(pad_value, float):
            pad_kwargs = {"constant_values": pad_value, "mode": "constant"}
        elif pad_value == "edge":
            pad_kwargs = {"mode": "edge"}
        else:
            print("Unrecognized pad value detected.")
        input_shape = data.shape
        target_image_shape = (input_shape[0], *target_shape)
        target_label_shape = (1, *target_shape)
        return input_shape, target_image_shape, target_label_shape, pad_kwargs

    def __croppad__(
        self,
        data_dict: np.ndarray,
        image_properties: dict,
        input_shape: np.ndarray,
        p_oversample_foreground: float,
        target_image_shape: list | tuple,
        target_label_shape: list | tuple,
        **pad_kwargs,
    ):
        image = data_dict[self.data_key]
        if data_dict.get(self.label_key) is not None:
            label = data_dict[self.label_key]
        else:
            label = None

        image, label = numpy_croppad(
            image=image,
            image_properties=image_properties,
            label=label,
            input_dims=len(input_shape[1:]),
            patch_size=self.patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )

        data_dict[self.data_key] = image
        if label is not None:
            data_dict[self.label_key] = label
        return data_dict

    def __call__(
        self, packed_data_dict=None, image_properties=None, **unpacked_data_dict
    ):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict

        input_shape, target_image_shape, target_label_shape, pad_kwargs = (
            self.get_params(
                data=data_dict[self.data_key],
                pad_value=self.pad_value,
                target_shape=self.patch_size,
            )
        )

        data_dict = self.__croppad__(
            data_dict=data_dict,
            image_properties=image_properties,
            input_shape=input_shape,
            p_oversample_foreground=self.p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
        return data_dict


class Torch_CropPad(BaseTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        pad_value: Union[Literal["min", "zero", "replicate"], int, float] = "min",
        patch_size: tuple | list = None,
        p_oversample_foreground=0.0,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.pad_value = pad_value
        self.patch_size = patch_size
        self.p_oversample_foreground = p_oversample_foreground

    @staticmethod
    def get_params(data, pad_value, target_shape):
        if pad_value == "min":
            pad_kwargs = {"value": data.min(), "mode": "constant"}
        elif pad_value == "zero":
            pad_kwargs = {"value": torch.zeros(1, dtype=data.dtype), "mode": "constant"}
        elif isinstance(pad_value, int) or isinstance(pad_value, float):
            pad_kwargs = {"value": pad_value, "mode": "constant"}
        elif pad_value == "replicate":
            pad_kwargs = {"mode": "replicate"}
        else:
            print("Unrecognized pad value detected.")
        input_shape = data.shape
        target_image_shape = (input_shape[0], *target_shape)
        target_label_shape = (1, *target_shape)
        return input_shape, target_image_shape, target_label_shape, pad_kwargs

    def __croppad__(
        self,
        data_dict: np.ndarray,
        foreground_locations: dict,
        input_shape: np.ndarray,
        p_oversample_foreground: float,
        target_image_shape: list | tuple,
        target_label_shape: list | tuple,
        **pad_kwargs,
    ):
        image = data_dict[self.data_key]
        if data_dict.get(self.label_key) is not None:
            label = data_dict[self.label_key]
        else:
            label = None

        image, label = torch_croppad(
            image=image,
            input_dims=len(input_shape[1:]),
            patch_size=self.patch_size,
            foreground_locations=foreground_locations,
            label=label,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )

        data_dict[self.data_key] = image
        if label is not None:
            data_dict[self.label_key] = label
        return data_dict

    def __call__(self, data_dict):
        input_shape, target_image_shape, target_label_shape, pad_kwargs = (
            self.get_params(
                data=data_dict[self.data_key],
                pad_value=self.pad_value,
                target_shape=self.patch_size,
            )
        )

        data_dict = self.__croppad__(
            data_dict=data_dict,
            foreground_locations=data_dict.get("foreground_locations"),
            input_shape=input_shape,
            p_oversample_foreground=self.p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
        return data_dict


class Numpy_Pad(Numpy_CropPad):
    def __call__(
        self, packed_data_dict=None, image_properties=None, **unpacked_data_dict
    ):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        image_shape = data_dict[self.data_key].shape
        original_patch_size = self.patch_size
        self.patch_size = fit_image_to_patch_size(self.patch_size, image_shape[1:])
        result = super().__call__(
            packed_data_dict=packed_data_dict,
            image_properties=image_properties,
            **unpacked_data_dict,
        )
        self.patch_size = (
            original_patch_size  # Reset patch size to original after padding
        )
        return result


class Torch_Pad(Torch_CropPad):
    def __call__(self, data_dict):
        original_patch_size = self.patch_size
        image_shape = data_dict[self.data_key].shape
        patch_size = fit_image_to_patch_size(original_patch_size, image_shape[1:])
        self.patch_size = patch_size
        result = super().__call__(data_dict)
        self.patch_size = original_patch_size
        return result


class Torch_CenterCrop(BaseTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        target_size: tuple | list = None,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.target_size = target_size

    @staticmethod
    def get_params():
        return

    def __centercrop__(self, data_dict):
        image = data_dict[self.data_key]

        if len(self.target_size) == 2:
            h, w = torch.tensor(image.shape[1:])
            h_delta, w_delta = torch.tensor(image.shape[1:]) - torch.tensor(
                self.target_size
            )
            h1, h2 = torch.floor(h_delta / 2).to(int), torch.ceil(h_delta / 2).to(int)
            w1, w2 = torch.floor(w_delta / 2).to(int), torch.ceil(w_delta / 2).to(int)
            slices = (slice(None, None), slice(h1, h - h2), slice(w1, w - w2))

        if len(self.target_size) == 3:
            h, w, d = torch.tensor(image.shape[1:])
            h_delta, w_delta, d_delta = torch.tensor(image.shape[1:]) - torch.tensor(
                self.target_size
            )
            h1, h2 = torch.floor(h_delta / 2).to(int), torch.ceil(h_delta / 2).to(int)
            w1, w2 = torch.floor(w_delta / 2).to(int), torch.ceil(w_delta / 2).to(int)
            d1, d2 = torch.floor(d_delta / 2).to(int), torch.ceil(d_delta / 2).to(int)
            slices = (
                slice(None, None),
                slice(h1, h - h2),
                slice(w1, w - w2),
                slice(d1, d - d2),
            )

        data_dict[self.data_key] = image[slices]
        if data_dict.get(self.label_key) is not None:
            data_dict[self.label_key] = data_dict[self.label_key][slices]

        return data_dict

    def __call__(self, data_dict):
        data_dict = self.__centercrop__(
            data_dict=data_dict,
        )
        return data_dict



import numpy as np
import torch


# Stolen from Batchgenerators to avoid import error caused by deprecated modules imported in
# Batchgenerators.
def numpy_gamma(
    data_sample,
    gamma_range=(0.5, 2),
    invert_image=False,
    epsilon=1e-7,
    per_channel=False,
    clip_to_input_range=False,
):
    if invert_image:
        data_sample = -data_sample

    if not per_channel:
        if np.random.random() < 0.5 and gamma_range[0] < 1:
            gamma = np.random.uniform(gamma_range[0], 1)
        else:
            gamma = np.random.uniform(max(gamma_range[0], 1), gamma_range[1])
        img_min = data_sample.min()
        img_max = data_sample.max()
        img_range = img_max - img_min
        data_sample = (
            np.power(((data_sample - img_min) / float(img_range + epsilon)), gamma)
            * img_range
            + img_min
        )
        if clip_to_input_range:
            data_sample = np.clip(data_sample, a_min=img_min, a_max=img_max)
    else:
        for c in range(data_sample.shape[0]):
            if np.random.random() < 0.5 and gamma_range[0] < 1:
                gamma = np.random.uniform(gamma_range[0], 1)
            else:
                gamma = np.random.uniform(max(gamma_range[0], 1), gamma_range[1])
            img_min = data_sample[c].min()
            img_max = data_sample[c].max()
            img_range = img_max - img_min
            data_sample[c] = (
                np.power(
                    ((data_sample[c] - img_min) / float(img_range + epsilon)), gamma
                )
                * float(img_range + epsilon)
                + img_min
            )
            if clip_to_input_range:
                data_sample[c] = np.clip(data_sample[c], a_min=img_min, a_max=img_max)
    if invert_image:
        data_sample = -data_sample
    return data_sample


def torch_gamma(
    image: torch.Tensor,
    gamma_range=(0.5, 2),
    invert_image=False,
    epsilon=1e-7,
    per_channel=False,
    clip_to_input_range=False,
) -> torch.Tensor:
    if invert_image:
        image = -image

    if not per_channel:
        if torch.rand(1).item() < 0.5 and gamma_range[0] < 1:
            gamma = torch.rand(1).item() * (1 - gamma_range[0]) + gamma_range[0]
        else:
            gamma = torch.rand(1).item() * (
                gamma_range[1] - max(gamma_range[0], 1)
            ) + max(gamma_range[0], 1)

        img_min = image.min()
        img_max = image.max()
        img_range = img_max - img_min

        image = (
            torch.pow(((image - img_min) / (img_range + epsilon)), gamma) * img_range
            + img_min
        )

        if clip_to_input_range:
            image = torch.clamp(image, min=img_min, max=img_max)
    else:
        for c in range(image.shape[0]):
            if torch.rand(1).item() < 0.5 and gamma_range[0] < 1:
                gamma = torch.rand(1).item() * (1 - gamma_range[0]) + gamma_range[0]
            else:
                gamma = torch.rand(1).item() * (
                    gamma_range[1] - max(gamma_range[0], 1)
                ) + max(gamma_range[0], 1)

            img_min = image[c].min()
            img_max = image[c].max()
            img_range = img_max - img_min

            image[c] = (
                torch.pow(((image[c] - img_min) / (img_range + epsilon)), gamma)
                * (img_range + epsilon)
                + img_min
            )

            if clip_to_input_range:
                image[c] = torch.clamp(image[c], min=img_min, max=img_max)

    return image



import torch
import torch.nn as nn
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.dropout import _DropoutNd
from typing import List, Tuple, Type, Union

__all__ = [
    "ConvDropoutNormNonlin",
    "ConvDropoutNorm",
    "DoubleConvDropoutNormNonlin",
    "DoubleLayerResBlock",
    "MultiLayerConvDropoutNormNonlin",
    "MultiLayerResBlock",
    "ResidualBlock",
    "StackedResidualBlocks",
]


class ConvDropoutNormNonlin(nn.Module):
    """
    2D Convolutional layers
    Arguments:
    num_in_filters {int} -- number of input filters
    num_out_filters {int} -- number of output filters
    kernel_size {tuple} -- size of the convolving kernel
    stride {tuple} -- stride of the convolution (default: {(1, 1)})
    activation {str} -- activation function (default: {'relu'})
    """

    def __init__(
        self,
        input_channels,
        output_channels,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        modules = []
        self.conv = self.conv_op(
            input_channels,
            output_channels,
            padding=conv_kwargs["kernel_size"] // 2,
            dilation=1,
            **self.conv_kwargs,
        )
        modules.append(self.conv)

        if self.dropout_op is not None:
            self.dropout = self.dropout_op(**self.dropout_op_kwargs)
            modules.append(self.dropout)
        if self.norm_op is not None:
            self.norm_op = self.norm_op(output_channels, **self.norm_op_kwargs)
            modules.append(self.norm_op)
        if self.nonlin is not None:
            self.nonlin = self.nonlin(**self.nonlin_kwargs)
            modules.append(self.nonlin)

        self.all_modules = nn.Sequential(*modules)

    def forward(self, x):
        return self.all_modules(x)


class ConvDropoutNorm(ConvDropoutNormNonlin):
    def forward(self, x):
        x = self.conv(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return self.norm(x)


class DoubleConvDropoutNormNonlin(nn.Module):
    """
    2/3D Convolutional layers
    Arguments:
    num_in_filters {int} -- number of input filters
    num_out_filters {int} -- number of output filters
    kernel_size {tuple} -- size of the convolving kernel
    stride {tuple} -- stride of the convolution (default: {(1, 1)})
    activation {str} -- activation function (default: {'relu'})
    """

    def __init__(
        self,
        input_channels,
        output_channels,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        self.conv1 = ConvDropoutNormNonlin(
            input_channels,
            output_channels,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )
        self.conv2 = ConvDropoutNormNonlin(
            output_channels,
            output_channels,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        return x


class MultiLayerConvDropoutNormNonlin(nn.Module):
    """
    2D Convolutional layers
    Arguments:
    num_in_filters {int} -- number of input filters
    num_out_filters {int} -- number of output filters
    num_layers {int} -- number of conv layers, must be at least 1
    kernel_size {tuple} -- size of the convolving kernel
    stride {tuple} -- stride of the convolution (default: {(1, 1)})
    activation {str} -- activation function (default: {'relu'})
    """

    def __init__(
        self,
        input_channels,
        output_channels,
        num_layers=2,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        assert num_layers >= 1, "Number of layers must be at least 1, got {}".format(
            num_layers
        )
        self.num_layers = num_layers

        self.conv1 = ConvDropoutNormNonlin(
            input_channels,
            output_channels,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        for layer in range(2, num_layers + 1):
            setattr(
                self,
                f"conv{layer}",
                ConvDropoutNormNonlin(
                    output_channels,
                    output_channels,
                    self.conv_op,
                    self.conv_kwargs,
                    self.norm_op,
                    self.norm_op_kwargs,
                    self.dropout_op,
                    self.dropout_op_kwargs,
                    self.nonlin,
                    self.nonlin_kwargs,
                ),
            )

    def forward(self, x):
        x = self.conv1(x)
        for layer in range(2, self.num_layers + 1):
            x = getattr(self, f"conv{layer}")(x)

        return x

    @staticmethod
    def get_block_constructor(n_layers):
        def _block(input_channels, output_channels, **kwargs):
            return MultiLayerConvDropoutNormNonlin(
                input_channels, output_channels, num_layers=n_layers, **kwargs
            )

        return _block


class DoubleLayerResBlock(nn.Module):
    """
    2D Convolutional layers
    Arguments:
    num_in_filters {int} -- number of input filters
    num_out_filters {int} -- number of output filters
    num_layers {int} -- number of conv layers, must be at least 1
    kernel_size {tuple} -- size of the convolving kernel
    stride {tuple} -- stride of the convolution (default: {(1, 1)})
    activation {str} -- activation function (default: {'relu'})
    """

    def __init__(
        self,
        input_channels,
        output_channels,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.0, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        assert conv_kwargs["dilation"] == 1, "Dilation must be 1 for residual blocks"

        self.conv1 = ConvDropoutNormNonlin(
            input_channels,
            output_channels,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        if (conv_kwargs["stride"] != 1) or (input_channels != output_channels):
            self.downsample_skip = nn.Sequential(
                conv_op(
                    input_channels,
                    output_channels,
                    kernel_size=1,
                    padding=0,
                    stride=conv_kwargs["stride"],
                    bias=False,
                ),
                norm_op(output_channels, **norm_op_kwargs),
            )
        else:
            self.downsample_skip = lambda x: x

        self.conv2 = ConvDropoutNorm(
            output_channels,
            output_channels,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.final_nonlin = self.nonlin(**self.nonlin_kwargs)

    def forward(self, x):
        residual = x

        x = self.conv1(x)
        x = self.conv2(x)

        x += self.downsample_skip(residual)
        x = self.final_nonlin(x)

        return x


class MultiLayerResBlock(nn.Module):
    """
    2D Convolutional layers
    Arguments:
    num_in_filters {int} -- number of input filters
    num_out_filters {int} -- number of output filters
    num_layers {int} -- number of conv layers, must be at least 1
    kernel_size {tuple} -- size of the convolving kernel
    stride {tuple} -- stride of the convolution (default: {(1, 1)})
    activation {str} -- activation function (default: {'relu'})
    """

    def __init__(
        self,
        input_channels,
        output_channels,
        num_layers=2,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        assert num_layers >= 1, "Number of layers must be at least 1, got {}".format(
            num_layers
        )

        assert conv_kwargs["stride"] == 1, "Stride must be 1 for residual blocks"
        assert conv_kwargs["conv_dilation"] == 1, (
            "Dilation must be 1 for residual blocks"
        )

        self.num_layers = num_layers

        self.conv1 = ConvDropoutNormNonlin(
            input_channels,
            output_channels,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        if (conv_kwargs["stride"] != 1) or (input_channels != output_channels):
            self.downsample_skip = nn.Sequential(
                conv_op(
                    input_channels,
                    output_channels,
                    kernel_size=1,
                    padding=0,
                    stride=conv_kwargs["stride"],
                    bias=False,
                ),
                norm_op(output_channels, **norm_op_kwargs),
            )
        else:
            self.downsample_skip = lambda x: x

        for layer in range(2, num_layers + 1):
            if layer < num_layers:
                setattr(
                    self,
                    f"conv{layer}",
                    ConvDropoutNormNonlin(
                        output_channels,
                        output_channels,
                        self.conv_op,
                        self.conv_kwargs,
                        self.norm_op,
                        self.norm_op_kwargs,
                        self.dropout_op,
                        self.dropout_op_kwargs,
                        self.nonlin,
                        self.nonlin_kwargs,
                    ),
                )
            else:
                # Last layer does not have activation, is added after residual
                setattr(
                    self,
                    f"conv{layer}",
                    ConvDropoutNorm(
                        output_channels,
                        output_channels,
                        self.conv_op,
                        self.conv_kwargs,
                        self.norm_op,
                        self.norm_op_kwargs,
                        self.dropout_op,
                        self.dropout_op_kwargs,
                        self.nonlin,
                        self.nonlin_kwargs,
                    ),
                )

        self.final_nonlin = self.nonlin(**self.nonlin_kwargs)

    def forward(self, x):
        residual = x
        x = self.conv1(x)
        for layer in range(2, self.num_layers + 1):
            x = getattr(self, f"conv{layer}")(x)

        x += self.downsample_skip(residual)
        x = self.final_nonlin(x)

        return x

    @staticmethod
    def get_block_constructor(n_layers):
        def _block(input_channels, output_channels, **kwargs):
            return MultiLayerResBlock(
                input_channels, output_channels, num_layers=n_layers, **kwargs
            )

        return _block


class ResidualBlock(nn.Module):
    def __init__(
        self,
        conv_op: Type[_ConvNd],
        pool_op: Union[nn.AvgPool2d, nn.AvgPool3d],
        input_channels: int,
        output_channels: int,
        kernel_size: int,
        stride: int,
        conv_bias: bool = True,
        norm_op: Union[None, Type[nn.Module]] = None,
        norm_op_kwargs: dict = None,
        dropout_op: Union[None, Type[_DropoutNd]] = None,
        dropout_op_kwargs: dict = None,
        nonlin: Union[None, Type[torch.nn.Module]] = None,
        nonlin_kwargs: dict = None,
    ):
        super().__init__()
        self.input_channels = input_channels
        self.output_channels = output_channels
        self.stride = stride

        self.conv1 = ConvDropoutNormNonlin(
            input_channels=input_channels,
            output_channels=output_channels,
            conv_op=conv_op,
            conv_kwargs={
                "stride": stride,
                "kernel_size": kernel_size,
                "bias": conv_bias,
            },
            norm_op=norm_op,
            norm_op_kwargs=norm_op_kwargs,
            dropout_op=dropout_op,
            dropout_op_kwargs=dropout_op_kwargs,
            nonlin=nonlin,
            nonlin_kwargs=nonlin_kwargs,
        )

        self.conv2 = ConvDropoutNormNonlin(
            input_channels=output_channels,
            output_channels=output_channels,
            conv_op=conv_op,
            conv_kwargs={"stride": 1, "kernel_size": kernel_size, "bias": conv_bias},
            norm_op=norm_op,
            norm_op_kwargs=norm_op_kwargs,
            dropout_op=None,
            dropout_op_kwargs=None,
            nonlin=None,
            nonlin_kwargs=None,
        )

        self.nonlin2 = nonlin(**nonlin_kwargs) if nonlin is not None else lambda x: x

        has_stride = stride != 1
        requires_projection = input_channels != output_channels

        if has_stride or requires_projection:
            ops = []
            if stride != 1:
                ops.append(pool_op(kernel_size=stride, stride=stride))
            if requires_projection:
                ops.append(
                    ConvDropoutNormNonlin(
                        input_channels=input_channels,
                        output_channels=output_channels,
                        conv_op=conv_op,
                        conv_kwargs={"stride": 1, "kernel_size": 1, "bias": False},
                        norm_op=norm_op,
                        norm_op_kwargs=norm_op_kwargs,
                        dropout_op=None,
                        dropout_op_kwargs=None,
                        nonlin=None,
                        nonlin_kwargs=None,
                    )
                )
            self.skip = nn.Sequential(*ops)
        else:
            self.skip = lambda x: x

    def forward(self, x):
        residual = self.skip(x)
        out = self.conv2(self.conv1(x))
        out += residual
        return self.nonlin2(out)


class StackedResidualBlocks(nn.Module):
    def __init__(
        self,
        n_blocks: int,
        conv_op: Type[_ConvNd],
        pool_op: Union[nn.AvgPool2d, nn.AvgPool3d],
        input_channels: int,
        output_channels: Union[int, List[int], Tuple[int, ...]],
        kernel_size: Union[int, List[int], Tuple[int, ...]],
        initial_stride: Union[int, List[int], Tuple[int, ...]],
        conv_bias: bool = False,
        norm_op: Union[None, Type[nn.Module]] = None,
        norm_op_kwargs: dict = None,
        dropout_op: Union[None, Type[_DropoutNd]] = None,
        dropout_op_kwargs: dict = None,
        nonlin: Union[None, Type[torch.nn.Module]] = None,
        nonlin_kwargs: dict = None,
        block: Type[ResidualBlock] = ResidualBlock,
    ):
        super().__init__()
        if not isinstance(output_channels, (tuple, list)):
            output_channels = [output_channels] * n_blocks

        blocks = nn.Sequential(
            block(
                conv_op=conv_op,
                input_channels=input_channels,
                output_channels=output_channels[0],
                kernel_size=kernel_size,
                stride=initial_stride,
                conv_bias=conv_bias,
                norm_op=norm_op,
                norm_op_kwargs=norm_op_kwargs,
                dropout_op=dropout_op,
                dropout_op_kwargs=dropout_op_kwargs,
                nonlin=nonlin,
                nonlin_kwargs=nonlin_kwargs,
                pool_op=pool_op,
            ),
            *[
                block(
                    conv_op=conv_op,
                    input_channels=output_channels[n - 1],
                    output_channels=output_channels[n],
                    kernel_size=kernel_size,
                    stride=1,
                    conv_bias=conv_bias,
                    norm_op=norm_op,
                    norm_op_kwargs=norm_op_kwargs,
                    dropout_op=dropout_op,
                    dropout_op_kwargs=dropout_op_kwargs,
                    nonlin=nonlin,
                    nonlin_kwargs=nonlin_kwargs,
                    pool_op=pool_op,
                )
                for n in range(1, n_blocks)
            ],
        )

        self.blocks = blocks
        self.output_channels = output_channels[-1]

    def forward(self, x):
        return self.blocks(x)



# FROM: https://github.com/MIC-DKFZ/dynamic-network-architectures/blob/main/dynamic_network_architectures/architectures/primus.py
# Copied here for debugging reasons
import torch
from gardening_tools.modules.networks.components.eva import Eva
from gardening_tools.modules.networks.components.layers import LayerNormNd
from gardening_tools.modules.networks.components.transformer import (
    PatchDecode,
    PatchEmbed,
)
from gardening_tools.modules.networks.components.weight_init import InitWeights_He
from einops import rearrange
from gardening_tools.modules.networks.BaseNet import BaseNet
from timm.layers import RotaryEmbeddingCat
from torch import nn
from typing import Tuple


class Primus(BaseNet):
    def __init__(
        self,
        input_channels: int,
        embed_dim: int,
        patch_embed_size: Tuple[int, ...],
        num_classes: int,
        eva_depth: int = 24,
        eva_numheads: int = 16,
        input_shape: Tuple[int, ...] = None,
        decoder_norm=LayerNormNd,
        decoder_act=nn.GELU,
        num_register_tokens: int = 0,
        use_rot_pos_emb: bool = True,
        use_abs_pos_embed: bool = True,
        mlp_ratio=4 * 2 / 3,
        drop_path_rate=0,  # drops computations (multihead attention, mlp), Implementation of \
        # scaling might be useless here because this is not batch normed
        patch_drop_rate: float = 0.0,  # drops input patches, may be used for MAE style pretraining
        proj_drop_rate: float = 0.0,  # drops out things related to the projection. \
        # That is in the MLP and at the end of EVA attention
        attn_drop_rate: float = 0.0,  # drops attention, meaning connections between patches may bebroken up at random
        rope_impl=RotaryEmbeddingCat,
        rope_kwargs=None,
        init_values=None,
        scale_attn_inner=False,
    ):
        """
        Architecture as proposed in the Primus paper (https://arxiv.org/pdf/2503.01835)
        `Primus: Enforcing Attention Usage for 3D Medical Image Segmentation`

        consists of simple patch_embedding, a EVA ViT encoder with a few adatptations and a simple patch decoder.
        """
        assert input_shape is not None
        assert len(input_shape) == 3, "Currently only 3d is supported"
        assert all([j % i == 0 for i, j in zip(patch_embed_size, input_shape)])

        super().__init__()

        self.stem_weight_name = (
            "encoder.proj.weight"  # used during weight loading of pt for ft
        )

        self.encoder = PatchEmbed(patch_embed_size, input_channels, embed_dim)
        self.decoder = PatchDecode(
            patch_embed_size,
            embed_dim,
            num_classes,
            norm=decoder_norm,
            activation=decoder_act,
        )

        # we need to compute the ref_feat_shape for eva
        self.eva = Eva(
            embed_dim=embed_dim,
            depth=eva_depth,
            num_heads=eva_numheads,
            ref_feat_shape=tuple(
                [i // ds for i, ds in zip(input_shape, patch_embed_size)]
            ),
            num_reg_tokens=num_register_tokens,
            use_rot_pos_emb=use_rot_pos_emb,
            use_abs_pos_emb=use_abs_pos_embed,
            mlp_ratio=mlp_ratio,
            drop_path_rate=drop_path_rate,
            patch_drop_rate=patch_drop_rate,
            proj_drop_rate=proj_drop_rate,
            attn_drop_rate=attn_drop_rate,
            rope_impl=rope_impl,
            rope_kwargs=rope_kwargs,
            init_values=init_values,
            scale_attn_inner=scale_attn_inner,
        )
        # self.mask_token =
        self.mask_token: torch.Tensor
        self.register_buffer("mask_token", torch.zeros(1, 1, embed_dim))

        if num_register_tokens > 0:
            self.register_tokens = (
                nn.Parameter(torch.zeros(1, num_register_tokens, embed_dim))
                if num_register_tokens
                else None
            )
            nn.init.normal_(self.register_tokens, std=1e-6)
        else:
            self.register_tokens = None

        self.num_classes = num_classes

        self.encoder.apply(InitWeights_He(1e-2))
        self.decoder.apply(InitWeights_He(1e-2))
        # eva has its own initialization

    def restore_full_sequence(self, x, keep_indices, num_patches):
        """
        Restore the full sequence by filling blanks with mask tokens and reordering.
        """
        if keep_indices is None:
            return x, None
        B, num_kept, C = x.shape
        device = x.device

        # Create mask tokens for missing patches
        num_masked = num_patches - num_kept
        mask_tokens = self.mask_token.repeat(B, num_masked, 1)

        # Prepare an empty tensor for the restored sequence
        restored = torch.zeros(B, num_patches, C, device=device)
        restored_mask = torch.zeros(B, num_patches, dtype=torch.bool, device=device)

        # Assign the kept patches and mask tokens in the correct positions
        for i in range(B):
            kept_pos = keep_indices[i]
            # masked_pos_prior = torch.tensor([j for j in range(num_patches) if j not in kept_pos], device=device)
            # replacement of list comprehension
            # kept_pos_tensor = torch.tensor(kept_pos, device=device)  # Ensure kept_pos is a tensor
            all_indices = torch.arange(
                num_patches, device=device
            )  # Create tensor of all indices
            mask = torch.ones(
                num_patches, device=device, dtype=torch.bool
            )  # Start with all True
            mask[kept_pos] = False  # Set kept positions to False
            masked_pos = all_indices[mask]  # Extract indices not in kept_pos

            restored[i, kept_pos] = x[i]
            restored[i, masked_pos] = mask_tokens[i, : len(masked_pos)]
            restored_mask[i, kept_pos] = True

        return (restored, restored_mask)

    def forward(self, x, ret_mask=False):
        FW, FH, FD = x.shape[2:]  # Full W , ...
        x = self.encoder(x)
        # last output of the encoder is the input to EVA
        B, C, W, H, D = x.shape
        num_patches = W * H * D

        x = rearrange(x, "b c w h d -> b (h w d) c")
        if self.register_tokens is not None:
            x = torch.cat(
                (
                    self.register_tokens.expand(x.shape[0], -1, -1),
                    x,
                ),
                dim=1,
            )
        x, keep_indices = self.eva(x)

        if self.register_tokens is not None:
            x = x[:, self.register_tokens.shape[1] :]  # Removes the register tokens
        # In-fill in-active patches with empty tokens
        restored_x, restoration_mask = self.restore_full_sequence(
            x, keep_indices, num_patches
        )
        x = rearrange(restored_x, "b (h w d) c -> b c w h d", h=H, w=W, d=D)
        if restoration_mask is not None:
            mask = rearrange(restoration_mask, "b (h w d) -> b w h d", h=H, w=W, d=D)
            full_mask = (
                mask.repeat_interleave(FW // W, dim=1)
                .repeat_interleave(FH // H, dim=2)
                .repeat_interleave(FD // D, dim=3)
            )
            full_mask = full_mask[
                :, None, ...
            ]  # Add channel dimension  # [B, 1, W, H, D]
        else:
            full_mask = None

        dec_out = self.decoder(x)
        if ret_mask:
            return dec_out, full_mask
        else:
            return dec_out

    def compute_conv_feature_map_size(self, input_size):
        raise NotImplementedError("yuck")



import numpy as np
import torch
import torch.nn.functional as F


def select_foreground_voxel_to_include(foreground_locations):
    if isinstance(foreground_locations, list):
        locidx = np.random.choice(len(foreground_locations))
        location = foreground_locations[locidx]
    elif isinstance(foreground_locations, dict):
        selected_class = np.random.choice(list(foreground_locations.keys()))
        locidx = np.random.choice(len(foreground_locations[selected_class]))
        location = foreground_locations[selected_class][locidx]
    return location


def fit_patch_size_to_image_size(patch_size, image_size):
    if len(patch_size) == 2 and len(image_size) == 3:
        patch_size = np.min([patch_size, image_size[1:]], axis=0)
    else:
        patch_size = np.min([patch_size, image_size], axis=0)

    return [int(32 * (i // 32)) for i in patch_size]


def fit_image_to_patch_size(patch_size, image_size):
    if len(patch_size) == 2 and len(image_size) == 3:
        return np.max([patch_size, image_size[1:]], axis=0)
    else:
        return np.max([patch_size, image_size], axis=0)


def numpy_croppad(
    image: np.ndarray,
    foreground_locations,
    input_dims: np.ndarray,
    patch_size,
    p_oversample_foreground,
    target_image_shape: list | tuple,
    target_label_shape: list | tuple,
    label: np.ndarray = None,
    **pad_kwargs,
):
    if len(patch_size) == 3:
        image, label = numpy_croppad_3D_case_from_3D(
            image=image,
            foreground_locations=foreground_locations,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
    elif len(patch_size) == 2 and input_dims == 3:
        image, label = numpy_croppad_2D_case_from_3D(
            image=image,
            foreground_locations=foreground_locations,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
    elif len(patch_size) == 2 and input_dims == 2:
        image, label = numpy_croppad_2D_case_from_2D(
            image=image,
            foreground_locations=foreground_locations,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )

    return image, label


def numpy_croppad_3D_case_from_3D(
    image,
    foreground_locations,
    label,
    patch_size,
    p_oversample_foreground,
    target_image_shape,
    target_label_shape,
    **pad_kwargs,
):
    image_out = np.zeros(target_image_shape)
    label_out = np.zeros(target_label_shape)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(3):
        if image.shape[d + 1] < patch_size[d]:
            to_pad += [patch_size[d] - image.shape[d + 1]]
        else:
            to_pad += [0]

    pad_lb_x = to_pad[0] // 2
    pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_y = to_pad[1] // 2
    pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2
    pad_lb_z = to_pad[2] // 2
    pad_ub_z = to_pad[2] // 2 + to_pad[2] % 2

    # This is where we should implement any patch selection biases.
    # The final patch excted after augmentation will always be the center of this patch
    # to avoid interpolation artifacts near the borders
    crop_start_idx = []
    if len(foreground_locations) == 0 or np.random.uniform() >= p_oversample_foreground:
        for d in range(3):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(image.shape[d + 1] - patch_size[d] + 1)
                ]
    else:
        location = select_foreground_voxel_to_include(foreground_locations)
        for d in range(3):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d] - patch_size[d]),
                        min(location[d], image.shape[d + 1] - patch_size[d]) + 1,
                    )
                ]

    image_out[
        :,
        :,
        :,
        :,
    ] = np.pad(
        image[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
            crop_start_idx[2] : crop_start_idx[2] + patch_size[2],
        ],
        ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
        **pad_kwargs,
    )
    if label is None:
        return image_out, None
    label_out[
        :,
        :,
        :,
        :,
    ] = np.pad(
        label[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
            crop_start_idx[2] : crop_start_idx[2] + patch_size[2],
        ],
        ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
    )
    return image_out, label_out


def numpy_croppad_2D_case_from_3D(
    image,
    foreground_locations,
    label,
    patch_size,
    p_oversample_foreground,
    target_image_shape,
    target_label_shape,
    **pad_kwargs,
):
    """
    The possible input for this can be 2D or 3D data.
    For 2D we want to pad or crop as necessary.
    For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
    then pad or crop as necessary.
    """
    image_out = np.zeros(target_image_shape)
    label_out = np.zeros(target_label_shape)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(2):
        if image.shape[d + 2] < patch_size[d]:
            to_pad += [patch_size[d] - image.shape[d + 2]]
        else:
            to_pad += [0]

    pad_lb_y = to_pad[0] // 2
    pad_ub_y = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_z = to_pad[1] // 2
    pad_ub_z = to_pad[1] // 2 + to_pad[1] % 2

    # This is where we should implement any patch selection biases.
    # The final patch extracted after augmentation will always be the center of this patch
    # as this is where augmentation-induced interpolation artefacts are least likely
    crop_start_idx = []
    if len(foreground_locations) == 0 or np.random.uniform() >= p_oversample_foreground:
        x_idx = np.random.randint(image.shape[1])
        for d in range(2):
            if image.shape[d + 2] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(image.shape[d + 2] - patch_size[d] + 1)
                ]
    else:
        location = select_foreground_voxel_to_include(foreground_locations)
        x_idx = location[0]
        for d in range(2):
            if image.shape[d + 2] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d + 1] - patch_size[d]),
                        min(location[d + 1], image.shape[d + 2] - patch_size[d]) + 1,
                    )
                ]

    image_out[:, :, :] = np.pad(
        image[
            :,
            x_idx,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        ((0, 0), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
        **pad_kwargs,
    )

    if label is None:
        return image_out, None

    label_out[:, :, :] = np.pad(
        label[
            :,
            x_idx,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        ((0, 0), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
    )

    return image_out, label_out


def numpy_croppad_2D_case_from_2D(
    image,
    foreground_locations,
    label,
    patch_size,
    p_oversample_foreground,
    target_image_shape,
    target_label_shape,
    **pad_kwargs,
):
    """
    The possible input for this can be 2D or 3D data.
    For 2D we want to pad or crop as necessary.
    For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
    then pad or crop as necessary.
    """
    image_out = np.zeros(target_image_shape)
    label_out = np.zeros(target_label_shape)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(2):
        if image.shape[d + 1] < patch_size[d]:
            to_pad += [patch_size[d] - image.shape[d + 1]]
        else:
            to_pad += [0]

    pad_lb_x = to_pad[0] // 2
    pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_y = to_pad[1] // 2
    pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2

    # This is where we should implement any patch selection biases.
    # The final patch extracted after augmentation will always be the center of this patch
    # as this is where artefacts are least present
    crop_start_idx = []
    if len(foreground_locations) == 0 or np.random.uniform() >= p_oversample_foreground:
        for d in range(2):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(image.shape[d + 1] - patch_size[d] + 1)
                ]
    else:
        location = select_foreground_voxel_to_include(foreground_locations)
        for d in range(2):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d] - patch_size[d]),
                        min(location[d], image.shape[d + 1] - patch_size[d]) + 1,
                    )
                ]

    image_out[:, :, :] = np.pad(
        image[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y)),
        **pad_kwargs,
    )

    if label is None:  # Reconstruction/inpainting
        return image_out, None

    if len(label.shape) == 1:  # Classification
        return image_out, label

    label_out[:, :, :] = np.pad(
        label[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y)),
    )

    return image_out, label_out


def torch_croppad(
    image: torch.tensor,
    patch_size,
    input_dims: torch.tensor,
    target_image_shape: list | tuple,
    target_label_shape: list | tuple,
    p_oversample_foreground=0.0,
    foreground_locations=None,
    label: torch.tensor = None,
    **pad_kwargs,
):
    """
    Crops and pads the input image and label to the specified patch size.
    Input includes channel/modality dimension so 3D is actually 4D and 2D is 3D.
    """
    if foreground_locations is None:
        foreground_locations = []

    if len(patch_size) == 3:
        image, label = torch_croppad_3D_case_from_3D(
            image=image,
            foreground_locations=foreground_locations,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
    elif len(patch_size) == 2 and input_dims == 3:
        image, label = torch_croppad_2D_case_from_3D(
            image=image,
            foreground_locations=foreground_locations,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
    elif len(patch_size) == 2 and input_dims == 2:
        image, label = torch_croppad_2D_case_from_2D(
            image=image,
            foreground_locations=foreground_locations,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )

    return image, label


def torch_croppad_3D_case_from_3D(
    image,
    foreground_locations,
    label,
    patch_size,
    p_oversample_foreground,
    target_image_shape,
    target_label_shape,
    **pad_kwargs,
):
    image_out = torch.zeros(target_image_shape, device=image.device)
    label_out = torch.zeros(target_label_shape, device=image.device)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(3):
        if image.shape[d + 1] < patch_size[d]:
            to_pad += [patch_size[d] - image.shape[d + 1]]
        else:
            to_pad += [0]

    pad_lb_x = to_pad[0] // 2
    pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_y = to_pad[1] // 2
    pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2
    pad_lb_z = to_pad[2] // 2
    pad_ub_z = to_pad[2] // 2 + to_pad[2] % 2

    # This is where we should implement any patch selection biases.
    # The final patch excted after augmentation will always be the center of this patch
    # to avoid interpolation artifacts near the borders
    crop_start_idx = []
    if len(foreground_locations) == 0 or np.random.uniform() >= p_oversample_foreground:
        for d in range(3):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(image.shape[d + 1] - patch_size[d] + 1)
                ]
    else:
        location = select_foreground_voxel_to_include(foreground_locations)
        for d in range(3):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d] - patch_size[d]),
                        min(location[d], image.shape[d + 1] - patch_size[d]) + 1,
                    )
                ]

    image_out[
        :,
        :,
        :,
        :,
    ] = F.pad(
        image[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
            crop_start_idx[2] : crop_start_idx[2] + patch_size[2],
        ],
        (
            pad_lb_z,
            pad_ub_z,
            pad_lb_y,
            pad_ub_y,
            pad_lb_x,
            pad_ub_x,
            0,
            0,
        ),
        **pad_kwargs,
    )
    if label is None:
        return image_out, None
    label_out[
        :,
        :,
        :,
        :,
    ] = F.pad(
        label[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
            crop_start_idx[2] : crop_start_idx[2] + patch_size[2],
        ],
        (
            pad_lb_z,
            pad_ub_z,
            pad_lb_y,
            pad_ub_y,
            pad_lb_x,
            pad_ub_x,
            0,
            0,
        ),
    )
    return image_out, label_out


def torch_croppad_2D_case_from_3D(
    image,
    foreground_locations,
    label,
    patch_size,
    p_oversample_foreground,
    target_image_shape,
    target_label_shape,
    **pad_kwargs,
):
    """
    The possible input for this can be 2D or 3D data.
    For 2D we want to pad or crop as necessary.
    For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
    then pad or crop as necessary.
    """
    image_out = torch.zeros(target_image_shape, device=image.device)
    label_out = torch.zeros(target_label_shape, device=image.device)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(2):
        if image.shape[d + 2] < patch_size[d]:
            to_pad += [patch_size[d] - image.shape[d + 2]]
        else:
            to_pad += [0]

    pad_lb_y = to_pad[0] // 2
    pad_ub_y = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_z = to_pad[1] // 2
    pad_ub_z = to_pad[1] // 2 + to_pad[1] % 2

    # This is where we should implement any patch selection biases.
    # The final patch extracted after augmentation will always be the center of this patch
    # as this is where augmentation-induced interpolation artefacts are least likely
    crop_start_idx = []
    if len(foreground_locations) == 0 or np.random.uniform() >= p_oversample_foreground:
        x_idx = np.random.randint(image.shape[1])
        for d in range(2):
            if image.shape[d + 2] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(image.shape[d + 2] - patch_size[d] + 1)
                ]
    else:
        location = select_foreground_voxel_to_include(foreground_locations)
        x_idx = location[0]
        for d in range(2):
            if image.shape[d + 2] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d + 1] - patch_size[d]),
                        min(location[d + 1], image.shape[d + 2] - patch_size[d]) + 1,
                    )
                ]

    image_out[:, :, :] = F.pad(
        image[
            :,
            x_idx,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        (
            pad_lb_z,
            pad_ub_z,
            pad_lb_y,
            pad_ub_y,
            0,
            0,
        ),
        **pad_kwargs,
    )

    if label is None:
        return image_out, None

    label_out[:, :, :] = F.pad(
        label[
            :,
            x_idx,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        (
            pad_lb_z,
            pad_ub_z,
            pad_lb_y,
            pad_ub_y,
            0,
            0,
        ),
    )

    return image_out, label_out


def torch_croppad_2D_case_from_2D(
    image,
    foreground_locations,
    label,
    patch_size,
    p_oversample_foreground,
    target_image_shape,
    target_label_shape,
    **pad_kwargs,
):
    """
    The possible input for this can be 2D or 3D data.
    For 2D we want to pad or crop as necessary.
    For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
    then pad or crop as necessary.
    """
    image_out = torch.zeros(target_image_shape, device=image.device)
    label_out = torch.zeros(target_label_shape, device=image.device)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(2):
        if image.shape[d + 1] < patch_size[d]:
            to_pad += [patch_size[d] - image.shape[d + 1]]
        else:
            to_pad += [0]

    pad_lb_x = to_pad[0] // 2
    pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_y = to_pad[1] // 2
    pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2

    # This is where we should implement any patch selection biases.
    # The final patch extracted after augmentation will always be the center of this patch
    # as this is where artefacts are least present
    crop_start_idx = []
    if len(foreground_locations) == 0 or np.random.uniform() >= p_oversample_foreground:
        for d in range(2):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(image.shape[d + 1] - patch_size[d] + 1)
                ]
    else:
        location = select_foreground_voxel_to_include(foreground_locations)
        for d in range(2):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d] - patch_size[d]),
                        min(location[d], image.shape[d + 1] - patch_size[d]) + 1,
                    )
                ]

    image_out[:, :, :] = F.pad(
        image[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        (
            pad_lb_y,
            pad_ub_y,
            pad_lb_x,
            pad_ub_x,
            0,
            0,
        ),
        **pad_kwargs,
    )

    if label is None:  # Reconstruction/inpainting
        return image_out, None

    if len(label.shape) == 1:  # Classification
        return image_out, label

    label_out[:, :, :] = F.pad(
        label[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        (
            pad_lb_y,
            pad_ub_y,
            pad_lb_x,
            pad_ub_x,
            0,
            0,
        ),
    )

    return image_out, label_out



import numpy as np
from gardening_tools.modules.transforms.BaseTransform import BaseTransform
from gardening_tools.functional.transforms.sampling import torch_resize
from gardening_tools.functional.transforms.sampling import (
    numpy_simulate_lowres,
    torch_simulate_lowres,
)


class Torch_Resize(BaseTransform):
    def __init__(
        self,
        data_key: str = "image",
        target_size: list = [],
        clip_to_input_range: bool = False,
    ):
        self.target_size = target_size
        self.data_key = data_key
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __resize__(self, data_dict):
        data_dict[self.data_key] = torch_resize(
            data_dict[self.data_key],
            target_size=self.target_size,
            clip_to_input_range=self.clip_to_input_range,
        )
        return data_dict

    def __call__(self, data_dict):
        data_dict = self.__resize__(data_dict)
        return data_dict


class Numpy_SimulateLowres(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        p_per_channel: float = 0.5,
        p_per_axis: float = 0.33,
        zoom_range=(0.5, 1.0),
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.p_per_channel = p_per_channel
        self.p_per_axis = p_per_axis
        self.zoom_range = zoom_range
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params(zoom_range, shape, p_per_axis):
        if isinstance(shape, (list, tuple)):
            shape = np.array(shape)
        zoom = np.random.uniform(*zoom_range)
        dim = len(shape)
        zoomed_shape = np.round(shape * zoom).astype(int)
        for i in range(dim):
            if np.random.uniform() < p_per_axis:
                shape[i] = zoomed_shape[i]
        return shape

    def __simulatelowres__(self, image, target_shape):
        image = numpy_simulate_lowres(
            image,
            target_shape=target_shape,
            clip_to_input_range=self.clip_to_input_range,
        )
        return image

    def __call__(self, data_dict):
        for b in range(data_dict[self.data_key].shape[0]):
            if np.random.uniform() < self.p_per_sample:
                for c in range(data_dict[self.data_key][b].shape[0]):
                    if np.random.uniform() < self.p_per_channel:
                        target_shape = self.get_params(
                            self.zoom_range,
                            data_dict[self.data_key][b, c].shape,
                            self.p_per_axis,
                        )
                        data_dict[self.data_key][b, c] = self.__simulatelowres__(
                            data_dict[self.data_key][b, c], target_shape
                        )
        return data_dict


class Torch_SimulateLowres(BaseTransform):
    def __init__(
        self,
        data_key="image",
        p_per_channel: float = 0.0,
        p_per_axis: float = 0.33,
        zoom_range=(0.5, 1.0),
        clip_to_input_range=False,
        batched: bool = True,
    ):
        self.data_key = data_key
        self.p_per_channel = p_per_channel
        self.p_per_axis = p_per_axis
        self.zoom_range = zoom_range
        self.clip_to_input_range = clip_to_input_range
        self.batched = batched

    @staticmethod
    def get_params(zoom_range, shape, p_per_axis):
        if isinstance(shape, (list, tuple)):
            shape = np.array(shape)
        zoom = np.random.uniform(*zoom_range)
        dim = len(shape)
        zoomed_shape = np.round(shape * zoom).astype(int)
        for i in range(dim):
            if np.random.uniform() < p_per_axis:
                shape[i] = zoomed_shape[i]
        return shape

    def __simulatelowres__(self, image, target_shape):
        image = torch_simulate_lowres(
            image,
            target_shape=target_shape,
            clip_to_input_range=self.clip_to_input_range,
        )
        return image

    def __call__(self, data_dict):
        if not self.batched:
            for c in range(data_dict[self.data_key].shape[0]):
                if np.random.uniform() < self.p_per_channel:
                    target_shape = self.get_params(
                        self.zoom_range,
                        data_dict[self.data_key][c].shape,
                        self.p_per_axis,
                    )
                    data_dict[self.data_key][c] = self.__simulatelowres__(
                        data_dict[self.data_key][c], target_shape
                    )
        else:
            for b in range(data_dict[self.data_key].shape[0]):
                for c in range(data_dict[self.data_key].shape[1]):
                    if np.random.uniform() < self.p_per_channel:
                        target_shape = self.get_params(
                            self.zoom_range,
                            data_dict[self.data_key][b, c].shape,
                            self.p_per_axis,
                        )
                        data_dict[self.data_key][b, c] = self.__simulatelowres__(
                            data_dict[self.data_key][b, c], target_shape
                        )

        return data_dict



import abc


class BaseTransform(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_params(self):
        """This will be called during __call__ and can be used to retrieve arbitrary parameters for the transform applied during __call__."""

    @abc.abstractmethod
    def __call__(self, dict):
        """
        This will be of the form __call__(self, dict):
        which allows calling it as either transform(dict)
        """



import logging
import torch
import torch.nn as nn
from gardening_tools.modules.networks.BaseNet import BaseNet
from gardening_tools.modules.networks.components.blocks import (
    ResidualBlock,
    MultiLayerConvDropoutNormNonlin,
)
from gardening_tools.modules.networks.components.encoders import ResidualUNetEncoder
from gardening_tools.modules.networks.components.decoders import ResidualUNetDecoder
from typing import List, Tuple, Type, Union


class ResidualEncoderUNet(BaseNet):
    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        dimensions: str,
        kernel_size: int,
        stride: int,
        features_per_stage: list,
        n_blocks_per_stage: Union[int, List[int], Tuple[int, ...]],
        n_conv_per_stage_decoder: Union[int, Tuple[int, ...], List[int]],
        conv_bias: bool = True,
        deep_supervision: bool = False,
        encoder_basic_block: Type[ResidualBlock] = ResidualBlock,
        decoder_basic_block: Type[
            MultiLayerConvDropoutNormNonlin
        ] = MultiLayerConvDropoutNormNonlin,
        norm_op_kwargs={"eps": 1e-05, "affine": True},
        dropout_op=None,
        dropout_op_kwargs=None,
        nonlin=torch.nn.LeakyReLU,
        nonlin_kwargs={"inplace": True},
        use_skip_connections: bool = True,
    ):
        super().__init__()

        if dimensions == "2D":
            conv_op = nn.Conv2d
            norm_op = nn.InstanceNorm2d
            upsample_op = torch.nn.ConvTranspose2d
            pool_op = nn.AvgPool2d
        elif dimensions == "3D":
            conv_op = nn.Conv3d
            norm_op = nn.InstanceNorm3d
            upsample_op = torch.nn.ConvTranspose3d
            pool_op = nn.AvgPool3d
        else:
            logging.warn("Uuh, dimensions not in ['2D', '3D']")

        self.num_classes = output_channels

        self.stem_weight_name = "encoder.stem.conv1.conv.weight"

        self.encoder = ResidualUNetEncoder(
            input_channels=input_channels,
            features_per_stage=features_per_stage,
            conv_op=conv_op,
            kernel_size=kernel_size,
            stride=stride,
            n_blocks_per_stage=n_blocks_per_stage,
            conv_bias=conv_bias,
            norm_op=norm_op,
            norm_op_kwargs=norm_op_kwargs,
            dropout_op=dropout_op,
            dropout_op_kwargs=dropout_op_kwargs,
            nonlin=nonlin,
            nonlin_kwargs=nonlin_kwargs,
            block=encoder_basic_block,
            pool_op=pool_op,
        )
        self.decoder = ResidualUNetDecoder(
            output_channels=output_channels,
            features_per_stage=features_per_stage[::-1],
            n_conv_per_stage=n_conv_per_stage_decoder,
            basic_block=decoder_basic_block,
            deep_supervision=deep_supervision,
            conv_op=conv_op,
            conv_kwargs={
                "kernel_size": kernel_size,
                "bias": conv_bias,
            },
            stride_for_transpose_conv=stride,
            upsample_op=upsample_op,
            norm_op=norm_op,
            norm_op_kwargs=norm_op_kwargs,
            nonlin=nonlin,
            nonlin_kwargs=nonlin_kwargs,
            use_skip_connections=use_skip_connections,
        )

    def forward(self, x):
        skips = self.encoder(x)
        return self.decoder(skips)



import numpy as np
import torch


def numpy_gibbs_ringing(image, num_sample, axis, clip_to_input_range: bool = False):
    img_min = image.min()
    img_max = image.max()
    m = min(0, img_min)
    image += abs(m)
    if len(image.shape) == 3:
        assert axis in [0, 1, 2], "Incorrect or no axis"

        h, w, d = image.shape
        if axis == 0:
            image = image.transpose(0, 2, 1)
            image = np.fft.fftshift(np.fft.fftn(image, s=[h, d, w]))
            image[:, :, 0 : int(np.ceil(w / 2) - np.ceil(num_sample / 2))] = 0
            image[:, :, int(np.ceil(w / 2) + np.ceil(num_sample / 2)) : w] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[h, d, w]))
            image = image.transpose(0, 2, 1)
        elif axis == 1:
            image = image.transpose(1, 2, 0)
            image = np.fft.fftshift(np.fft.fftn(image, s=[w, d, h]))
            image[:, :, 0 : int(np.ceil(h / 2) - np.ceil(num_sample / 2))] = 0
            image[:, :, int(np.ceil(h / 2) + np.ceil(num_sample / 2)) : h] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[w, d, h]))
            image = image.transpose(2, 0, 1)
        else:
            image = np.fft.fftshift(np.fft.fftn(image, s=[h, w, d]))
            image[:, :, 0 : int(np.ceil(d / 2) - np.ceil(num_sample / 2))] = 0
            image[:, :, int(np.ceil(d / 2) + np.ceil(num_sample / 2)) : d] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[h, w, d]))
    elif len(image.shape) == 2:
        assert axis in [0, 1], "incorrect or no axis"
        h, w = image.shape
        if axis == 0:
            image = np.fft.fftshift(np.fft.fftn(image, s=[h, w]))
            image[:, 0 : int(np.ceil(w / 2) - np.ceil(num_sample / 2))] = 0
            image[:, int(np.ceil(w / 2) + np.ceil(num_sample / 2)) : w] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[h, w]))
        else:
            image = image.conj().T
            image = np.fft.fftshift(np.fft.fftn(image, s=[w, h]))
            image[:, 0 : int(np.ceil(h / 2) - np.ceil(num_sample / 2))] = 0
            image[:, int(np.ceil(h / 2) + np.ceil(num_sample / 2)) : h] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[w, h]))
            image = image.conj().T
    image -= abs(m)
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)
    return image


def torch_gibbs_ringing(
    image: torch.Tensor,
    num_sample: int,
    mode: str = "rect",
    axes: list[int] = None,
    clip_to_input_range: bool = False,
) -> torch.Tensor:
    assert image.ndim in [2, 3], "Only 2D or 3D images supported"
    if mode == "rect":
        assert axes is not None and all(0 <= ax < image.ndim for ax in axes), (
            "Invalid axes for mode 'rect'"
        )

    img_min = image.min()
    img_max = image.max()
    offset = -img_min if img_min < 0 else 0
    image = image + offset

    kspace = torch.fft.fftshift(
        torch.fft.fftn(image, dim=list(range(image.ndim))), dim=list(range(image.ndim))
    )

    shape = image.shape
    center = [s // 2 for s in shape]

    if mode == "rect":
        mask = torch.ones_like(kspace, dtype=torch.bool)
        for axis in axes:
            c = center[axis]
            half = num_sample // 2
            slc = [slice(None)] * image.ndim
            for i in range(shape[axis]):
                if not (c - half <= i < c + half):
                    slc[axis] = slice(i, i + 1)
                    mask[tuple(slc)] = False
        kspace[~mask] = 0
    elif mode == "radial":
        coords = torch.meshgrid(
            [torch.arange(s, device=image.device) - c for s, c in zip(shape, center)],
            indexing="ij",
        )
        dist = torch.sqrt(sum((g.float() ** 2 for g in coords)))
        mask = dist <= num_sample
        kspace[~mask] = 0
    else:
        raise ValueError(f"Unsupported mode: {mode}")

    result = torch.fft.ifftn(
        torch.fft.ifftshift(kspace, dim=list(range(image.ndim))),
        dim=list(range(image.ndim)),
    )
    result = result.abs()

    result = result - offset
    if clip_to_input_range:
        result = torch.clamp(result, min=img_min, max=img_max)

    return result



from gardening_tools.modules.transforms.BaseTransform import BaseTransform


class CopyImageToLabel(BaseTransform):
    def __init__(self, copy=False, data_key="image", label_key="label"):
        self.copy = copy
        self.data_key = data_key
        self.label_key = label_key

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __copy__(self, data_dict):
        data_dict[self.label_key] = data_dict[self.data_key].copy()
        return data_dict

    def __call__(self, data_dict):
        if self.copy:
            data_dict = self.__copy__(data_dict)
        return data_dict


class Torch_CopyImageToLabel(BaseTransform):
    def __init__(self, copy=False, data_key="image", label_key="label"):
        self.copy = copy
        self.data_key = data_key
        self.label_key = label_key

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __copy__(self, data_dict):
        data_dict[self.label_key] = data_dict[self.data_key].detach().clone()
        return data_dict

    def __call__(self, data_dict):
        if self.copy:
            data_dict = self.__copy__(data_dict)
        return data_dict



import numpy as np


def verify_labels_are_equal(
    expected_labels: np.ndarray, actual_labels: np.ndarray, id=""
):
    if expected_labels.dtype == actual_labels.dtype:
        if np.all(np.isin(actual_labels, expected_labels)):
            return True
        else:
            print(
                f"Unexpected labels found for {id} \n"
                f"expected: {expected_labels} \n"
                f"found: {actual_labels}"
            )
            return False
    else:
        print(
            "make sure reference and target is the same dtype before comparing the labels. \n"
            f"reference is: {expected_labels.dtype} and target is: {actual_labels.dtype}"
        )
        return False


def verify_shapes_are_equal(reference_shape, target_shape, id=""):
    if np.all(reference_shape == target_shape):
        return True
    else:
        print(
            f"Sizes do not match for {id}"
            f"Image is: {reference_shape} while the label is {target_shape}"
        )
        return False



import torch
import torch.nn as nn
from gardening_tools.modules.networks.components.blocks import (
    DoubleConvDropoutNormNonlin,
)


class UNetDecoder(nn.Module):
    def __init__(
        self,
        output_channels: int = 1,
        starting_filters: int = 64,
        basic_block=DoubleConvDropoutNormNonlin,
        conv_op=nn.Conv3d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "bias": True,
        },
        deep_supervision=False,
        dropout_in_decoder=False,
        norm_op=nn.InstanceNorm3d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout3d,
        dropout_op_kwargs={"p": 0.0, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
        upsample_op=torch.nn.ConvTranspose3d,
        use_skip_connections=True,
        weightInitializer=None,
    ) -> None:
        super().__init__()

        # Task specific
        self.num_classes = output_channels
        self.filters = starting_filters

        # Model parameters
        self.basic_block = basic_block
        self.conv_op = conv_op
        self.conv_kwargs = conv_kwargs
        self.deep_supervision = deep_supervision
        self.norm_op_kwargs = norm_op_kwargs
        self.norm_op = norm_op
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.weightInitializer = weightInitializer
        self.use_skip_connections = use_skip_connections
        self.upsample = upsample_op

        self.upsample1 = self.upsample(
            self.filters * 16, self.filters * 8, kernel_size=2, stride=2
        )
        self.decoder_conv1 = self.basic_block(
            input_channels=self.filters * (16 if self.use_skip_connections else 8),
            output_channels=self.filters * 8,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.upsample2 = self.upsample(
            self.filters * 8, self.filters * 4, kernel_size=2, stride=2
        )
        self.decoder_conv2 = self.basic_block(
            input_channels=self.filters * (8 if self.use_skip_connections else 4),
            output_channels=self.filters * 4,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.upsample3 = self.upsample(
            self.filters * 4, self.filters * 2, kernel_size=2, stride=2
        )
        self.decoder_conv3 = self.basic_block(
            input_channels=self.filters * (4 if self.use_skip_connections else 2),
            output_channels=self.filters * 2,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.upsample4 = self.upsample(
            self.filters * 2, self.filters, kernel_size=2, stride=2
        )
        self.decoder_conv4 = self.basic_block(
            input_channels=self.filters * (2 if self.use_skip_connections else 1),
            output_channels=self.filters,
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.out_conv = self.conv_op(self.filters, self.num_classes, kernel_size=1)

        if self.deep_supervision:
            self.ds_out_conv0 = self.conv_op(
                self.filters * 16, self.num_classes, kernel_size=1
            )
            self.ds_out_conv1 = self.conv_op(
                self.filters * 8, self.num_classes, kernel_size=1
            )
            self.ds_out_conv2 = self.conv_op(
                self.filters * 4, self.num_classes, kernel_size=1
            )
            self.ds_out_conv3 = self.conv_op(
                self.filters * 2, self.num_classes, kernel_size=1
            )

        if self.weightInitializer is not None:
            print("initializing weights")
            self.apply(self.weightInitializer)

    def forward(self, xs):
        x_enc = xs[4]

        if self.use_skip_connections:
            x5 = torch.cat([self.upsample1(x_enc), xs[3]], dim=1)
            x5 = self.decoder_conv1(x5)

            x6 = torch.cat([self.upsample2(x5), xs[2]], dim=1)
            x6 = self.decoder_conv2(x6)

            x7 = torch.cat([self.upsample3(x6), xs[1]], dim=1)
            x7 = self.decoder_conv3(x7)

            x8 = torch.cat([self.upsample4(x7), xs[0]], dim=1)
            x8 = self.decoder_conv4(x8)
        else:
            x5 = self.decoder_conv1(self.upsample1(x_enc))
            x6 = self.decoder_conv2(self.upsample2(x5))
            x7 = self.decoder_conv3(self.upsample3(x6))
            x8 = self.decoder_conv4(self.upsample4(x7))

        # We only want to do multiple outputs during training, therefore it is only enabled
        # when grad is also enabled because that means we're training. And if for some reason
        # grad is enabled and you're not training, then there's other, bigger problems.
        if self.deep_supervision and torch.is_grad_enabled():
            ds0 = self.ds_out_conv0(xs[4])
            ds1 = self.ds_out_conv1(x5)
            ds2 = self.ds_out_conv2(x6)
            ds3 = self.ds_out_conv3(x7)
            ds4 = self.out_conv(x8)
            return [ds4, ds3, ds2, ds1, ds0]

        logits = self.out_conv(x8)

        return logits


class ResidualUNetDecoder(nn.Module):
    def __init__(
        self,
        output_channels: int = 1,
        features_per_stage: list = [],
        n_conv_per_stage: list = [],
        basic_block=DoubleConvDropoutNormNonlin,
        conv_op=nn.Conv3d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "bias": True,
        },
        deep_supervision=False,
        norm_op=nn.InstanceNorm3d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout3d,
        dropout_op_kwargs={"p": 0.0, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"inplace": True},
        stride_for_transpose_conv: int = 2,
        upsample_op=torch.nn.ConvTranspose3d,
        use_skip_connections=True,
        weightInitializer=None,
    ) -> None:
        super().__init__()

        # Task specific
        self.num_classes = output_channels
        self.features_per_stage = features_per_stage
        self.n_conv_per_stage = n_conv_per_stage

        # Model parameters
        self.basic_block = basic_block
        self.conv_op = conv_op
        self.conv_kwargs = conv_kwargs
        self.deep_supervision = deep_supervision
        self.norm_op_kwargs = norm_op_kwargs
        self.norm_op = norm_op
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.weightInitializer = weightInitializer
        self.use_skip_connections = use_skip_connections
        self.upsample = upsample_op

        self.upsample1 = self.upsample(
            self.features_per_stage[0],
            self.features_per_stage[1],
            kernel_size=stride_for_transpose_conv,
            stride=stride_for_transpose_conv,
        )

        self.decoder_conv1 = self.basic_block(
            input_channels=self.features_per_stage[1]
            * (2 if self.use_skip_connections else 1),
            output_channels=self.features_per_stage[1],
            num_layers=self.n_conv_per_stage[0],
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.upsample2 = self.upsample(
            self.features_per_stage[1],
            self.features_per_stage[2],
            kernel_size=stride_for_transpose_conv,
            stride=stride_for_transpose_conv,
        )
        self.decoder_conv2 = self.basic_block(
            input_channels=self.features_per_stage[2]
            * (2 if self.use_skip_connections else 1),
            output_channels=self.features_per_stage[2],
            num_layers=self.n_conv_per_stage[1],
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.upsample3 = self.upsample(
            self.features_per_stage[2],
            self.features_per_stage[3],
            kernel_size=stride_for_transpose_conv,
            stride=stride_for_transpose_conv,
        )
        self.decoder_conv3 = self.basic_block(
            input_channels=self.features_per_stage[3]
            * (2 if self.use_skip_connections else 1),
            output_channels=self.features_per_stage[3],
            num_layers=self.n_conv_per_stage[2],
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.upsample4 = self.upsample(
            self.features_per_stage[3],
            self.features_per_stage[4],
            kernel_size=stride_for_transpose_conv,
            stride=stride_for_transpose_conv,
        )
        self.decoder_conv4 = self.basic_block(
            input_channels=self.features_per_stage[4]
            * (2 if self.use_skip_connections else 1),
            output_channels=self.features_per_stage[4],
            num_layers=self.n_conv_per_stage[3],
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.upsample5 = self.upsample(
            self.features_per_stage[4],
            self.features_per_stage[5],
            kernel_size=stride_for_transpose_conv,
            stride=stride_for_transpose_conv,
        )

        self.decoder_conv5 = self.basic_block(
            input_channels=self.features_per_stage[5]
            * (2 if self.use_skip_connections else 1),
            output_channels=self.features_per_stage[5],
            num_layers=self.n_conv_per_stage[4],
            conv_op=self.conv_op,
            conv_kwargs=self.conv_kwargs,
            norm_op=self.norm_op,
            norm_op_kwargs=self.norm_op_kwargs,
            dropout_op=self.dropout_op,
            dropout_op_kwargs=self.dropout_op_kwargs,
            nonlin=self.nonlin,
            nonlin_kwargs=self.nonlin_kwargs,
        )

        self.out_conv = self.conv_op(
            self.features_per_stage[5], self.num_classes, kernel_size=1
        )

        if self.deep_supervision:
            self.ds_out_conv0 = self.conv_op(
                self.features_per_stage[0], self.num_classes, kernel_size=1
            )
            self.ds_out_conv1 = self.conv_op(
                self.features_per_stage[1], self.num_classes, kernel_size=1
            )
            self.ds_out_conv2 = self.conv_op(
                self.features_per_stage[2], self.num_classes, kernel_size=1
            )
            self.ds_out_conv3 = self.conv_op(
                self.features_per_stage[3], self.num_classes, kernel_size=1
            )
            self.ds_out_conv4 = self.conv_op(
                self.features_per_stage[4], self.num_classes, kernel_size=1
            )

        if self.weightInitializer is not None:
            print("initializing weights")
            self.apply(self.weightInitializer)

    def forward(self, xs):
        x_enc = xs[5]

        if self.use_skip_connections:
            x5 = torch.cat([self.upsample1(x_enc), xs[4]], dim=1)
            x5 = self.decoder_conv1(x5)

            x6 = torch.cat([self.upsample2(x5), xs[3]], dim=1)
            x6 = self.decoder_conv2(x6)

            x7 = torch.cat([self.upsample3(x6), xs[2]], dim=1)
            x7 = self.decoder_conv3(x7)

            x8 = torch.cat([self.upsample4(x7), xs[1]], dim=1)
            x8 = self.decoder_conv4(x8)

            x9 = torch.cat([self.upsample5(x8), xs[0]], dim=1)
            x9 = self.decoder_conv5(x9)
        else:
            x5 = self.decoder_conv1(self.upsample1(x_enc))
            x6 = self.decoder_conv2(self.upsample2(x5))
            x7 = self.decoder_conv3(self.upsample3(x6))
            x8 = self.decoder_conv4(self.upsample4(x7))
            x9 = self.decoder_conv5(self.upsample5(x8))

        # We only want to do multiple outputs during training, therefore it is only enabled
        # when grad is also enabled because that means we're training. And if for some reason
        # grad is enabled and you're not training, then there's other, bigger problems.
        if self.deep_supervision and torch.is_grad_enabled():
            ds0 = self.ds_out_conv0(xs[4])
            ds1 = self.ds_out_conv1(x5)
            ds2 = self.ds_out_conv2(x6)
            ds3 = self.ds_out_conv3(x7)
            ds4 = self.ds_out_conv4(x8)
            ds5 = self.out_conv(x9)
            return [ds5, ds4, ds3, ds2, ds1, ds0]

        logits = self.out_conv(x9)

        return logits



import torch
from gardening_tools.modules.transforms.BaseTransform import BaseTransform


class Torch_Mirror(BaseTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        p_per_sample: float = 1.0,
        axes=(0, 1, 2),
        p_mirror_per_axis=0.33,
        skip_label=False,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.p_per_sample = p_per_sample
        self.p_mirror_per_axis = p_mirror_per_axis
        self.axes = axes
        self.skip_label = skip_label

    @staticmethod
    def get_params():
        pass

    def __mirror__(self, data_dict, axes):
        image = data_dict[self.data_key]
        label = data_dict.get(self.label_key)

        if torch.rand(1).item() < self.p_per_sample:
            for axis in axes:
                if torch.rand(1).item() < self.p_mirror_per_axis:
                    flip_dim = axis + 1  # C=0, so X,Y,Z → dims 1,2,3
                    image = torch.flip(image, [flip_dim])
                    if label is not None and not self.skip_label:
                        label = torch.flip(label, [flip_dim])

        data_dict[self.data_key] = image
        if label is not None and not self.skip_label:
            data_dict[self.label_key] = label
        return data_dict

    def __call__(self, data_dict):
        return self.__mirror__(data_dict, self.axes)



import numpy as np
import torch
from gardening_tools.modules.networks.components.layers import LayerNormNd
from gardening_tools.modules.networks.components.utils import convert_dim_to_conv_op
from torch import nn
from typing import Tuple


class PatchEmbed(nn.Module):
    """
    Image to Patch Embedding.
    Loosely inspired by https://github.com/facebookresearch/segment-anything/blob/main/segment_anything/modeling/image_encoder.py#L364

    """

    def __init__(
        self,
        patch_size: Tuple[int, ...] = (16, 16, 16),
        input_channels: int = 3,
        embed_dim: int = 768,
    ) -> None:
        """
        Args:
            patch_size (Tuple): patch size.
            padding (Tuple): padding size of the projection layer.
            input_channels (int): Number of input image channels.
            embed_dim (int): Patch embedding dimension.
        """
        super().__init__()

        self.proj = convert_dim_to_conv_op(len(patch_size))(
            input_channels,
            embed_dim,
            kernel_size=patch_size,
            stride=patch_size,
            padding=0,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        returns shape (B, embed_dim, px, py, pz) where (px, py, pz) is patch_size.
        This output will need to be rearranged to whatever your transformer expects!
        """
        x = self.proj(x)
        return x


class PatchDecode(nn.Module):
    """
    Loosely inspired by SAM decoder
    https://github.com/facebookresearch/segment-anything/blob/main/segment_anything/modeling/mask_decoder.py#L53
    """

    def __init__(
        self,
        patch_size,
        embed_dim: int,
        out_channels: int,
        norm=LayerNormNd,
        activation=nn.GELU,
    ):
        """
        patch size must be 2^x, so 2, 4, 8, 16, 32, etc. Otherwise we die
        """
        super().__init__()

        def _round_to_8(inp):
            return int(max(8, np.round((inp + 1e-6) / 8) * 8))

        self.num_classes = out_channels

        num_stages = int(np.log(max(patch_size)) / np.log(2))
        strides = [
            [2 if (p / 2**n) % 2 == 0 else 1 for p in patch_size]
            for n in range(num_stages)
        ][::-1]
        dim_red = (embed_dim / (2 * out_channels)) ** (1 / num_stages)

        # don't question me
        channels = [embed_dim] + [
            _round_to_8(embed_dim / dim_red ** (x + 1)) for x in range(num_stages)
        ]
        channels[-1] = out_channels

        stages = []
        for s in range(num_stages - 1):
            stages.append(
                nn.Sequential(
                    nn.ConvTranspose3d(
                        channels[s],
                        channels[s + 1],
                        kernel_size=strides[s],
                        stride=strides[s],
                    ),
                    norm(channels[s + 1]),
                    activation(),
                )
            )
        stages.append(
            nn.ConvTranspose3d(
                channels[-2], channels[-1], kernel_size=strides[-1], stride=strides[-1]
            )
        )
        self.decode = nn.Sequential(*stages)

    def forward(self, x):
        """
        Expects input of shape (B, embed_dim, px, py, pz)! This will require you to reshape the output of your transformer!
        """
        return self.decode(x)



import os
from typing import Optional, List


def subdirs(
    folder: str,
    join: bool = True,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
    sort: bool = True,
) -> List[str]:
    """
    implementation by: https://github.com/MIC-DKFZ/batchgenerators
    """
    subdirectories = []
    with os.scandir(folder) as entries:
        for entry in entries:
            if (
                entry.is_dir()
                and (prefix is None or entry.name.startswith(prefix))
                and (suffix is None or entry.name.endswith(suffix))
            ):
                dir_path = entry.path if join else entry.name
                subdirectories.append(dir_path)

    if sort:
        subdirectories.sort()
    return subdirectories


def subfiles(
    folder: str,
    join: bool = True,
    prefix: str = None,
    suffix: str = None,
    sort: bool = True,
) -> List[str]:
    if join:
        res = [
            os.path.join(folder, i)
            for i in os.listdir(folder)
            if os.path.isfile(os.path.join(folder, i))
            and (prefix is None or i.startswith(prefix))
            and (suffix is None or i.endswith(suffix))
        ]
    else:
        res = [
            i
            for i in os.listdir(folder)
            if os.path.isfile(os.path.join(folder, i))
            and (prefix is None or i.startswith(prefix))
            and (suffix is None or i.endswith(suffix))
        ]
    if sort:
        res.sort()
    return res



from torch import nn
from torch.nn.modules.conv import _ConvNd
from typing import Type


def convert_dim_to_conv_op(dimension: int) -> Type[_ConvNd]:
    """
    :param dimension: 1, 2 or 3
    :return: conv Class of corresponding dimension
    """
    if dimension == 1:
        return nn.Conv1d
    elif dimension == 2:
        return nn.Conv2d
    elif dimension == 3:
        return nn.Conv3d
    else:
        raise ValueError("Unknown dimension. Only 1, 2 and 3 are supported")



import numpy as np
import nibabel as nib
from typing import Union


def nifti_or_np_to_np(array: Union[np.ndarray, nib.Nifti1Image]) -> np.ndarray:
    if isinstance(array, np.ndarray):
        return array
    if isinstance(array, (nib.Nifti1Image, nib.minc1.Minc1Image)):
        return array.get_fdata().astype(np.float32)
    else:
        raise TypeError(
            f"File data type invalid. Found: {type(array)} and expected nib.Nifti1Image or np.ndarray"
        )
